import numpy as np


def tetrahedral_mesh_aabbs(tetrahedra_points):
    """Compute axis-aligned bounding boxes of tetrahedra.

    Parameters
    ----------
    tetrahedra_points : array, shape (n_tetrahedra, 4, 3)
        Points that form the tetrahedra.

    Returns
    -------
    aabbs : array, shape (n_tetrahedra, 3, 2)
        Axis-aligned bounding boxes of tetrahedra.
    """
    mins = np.min(tetrahedra_points, axis=1)
    maxs = np.max(tetrahedra_points, axis=1)
    aabbs = np.dstack((mins, maxs))
    return aabbs


def center_of_mass_tetrahedral_mesh(tetrahedra_points):
    """Compute center of mass of a tetrahedral mesh.

    Assumes uniform density.

    Parameters
    ----------
    tetrahedra_points : array, shape (n_tetrahedra, 4, 3)
        Points that form the tetrahedra.

    Returns
    -------
    com : array, shape (3,)
        Center of mass.
    """
    volumes = tetrahedral_mesh_volumes(tetrahedra_points)
    centers = tetrahedra_points.mean(axis=1)
    return np.dot(volumes, centers) / np.sum(volumes)


def tetrahedral_mesh_volumes(tetrahedra_points):
    """Compute volumes of tetrahedra.

    Parameters
    ----------
    tetrahedra_points : array, shape (n_tetrahedra, 4, 3)
        Points that form the tetrahedra.

    Returns
    -------
    volumes : array, shape (n_tetrahedra,)
        Volumes of tetrahedra.
    """
    tetrahedra_edges = (tetrahedra_points[:, 1:]
                        - tetrahedra_points[:, np.newaxis, 0])
    return np.abs(np.sum(
        np.cross(tetrahedra_edges[:, 0], tetrahedra_edges[:, 1])
        * tetrahedra_edges[:, 2], axis=1)) / 6.0



"""Gilbert-Johnson-Keerthi (GJK) for distance calculation of convex shapes.

The GJK algorithm only works for convex shapes. Concave objects have to be
decomposed into convex shapes first.

This module contains several flavours of the GJK algorithm. Some of them
only detect collisions (intersections) and some of them calculate distances
between separated objects as well.

The original publication describing the algorithm is:

E.G. Gilbert, D.W. Johnson, S.S. Keerthi: A fast procedure for computing
the distance between complex objects in three-dimensional space, IEEE
Journal on Robotics and Automation (1988),
https://graphics.stanford.edu/courses/cs448b-00-winter/papers/gilbert.pdf
"""
from ._gjk_original import gjk_distance_original
from ._gjk_libccd import gjk_intersection_libccd
from ._gjk_jolt import gjk_distance_jolt, gjk_intersection_jolt


# Aliases
gjk = gjk_distance_jolt
gjk_distance = gjk_distance_jolt
gjk_intersection = gjk_intersection_jolt


__all__ = [
    "gjk_distance_original",
    "gjk_intersection_libccd",
    "gjk_distance_jolt",
    "gjk_intersection_jolt",
    "gjk_distance",
    "gjk_intersection",
    "gjk"
]



"""Utility functions."""
import math
import numba
import numpy as np


MAX_FLOAT = np.finfo(float).max
EPSILON = np.finfo(float).eps
HALF_PI = 0.5 * np.pi


@numba.njit(numba.float64[::1](numba.float64[::1]), cache=True)
def norm_vector(v):
    """Normalize vector.

    Parameters
    ----------
    v : array, shape (n,)
        nd vector

    Returns
    -------
    u : array, shape (n,)
        nd unit vector with norm 1 or the zero vector
    """
    norm = np.linalg.norm(v)
    if norm == 0.0:
        return v

    return v / norm


@numba.njit(cache=True)
def scalar_triple_product(a, b, c):
    r"""Scalar triple product: :math:`a \cdot (b \cross c)`.

    Also known as triple scalar product or box product and written as
    :math:`\left[ a b c \right]`.

    The value of the scalar triple product is the signed volume of a
    parallelepiped formed by the three independent vectors a, b, c. It is
    six times the volume of a tetrahedron spanned by these three vectors.

    The cross and dot product can be interchanged without changing the result:
    :math:`a \cdot (b \cross c) = (a \cross b) \cdot c`. The result also
    remains constant under cyclic permutation of the arguments:
    :math:`(a \cross b) \cdot c = (b \cross c) \cdot a = (c \cross a) \cdot c`.

    The scalar triple product is also the determinant of the matrix
    :math:`\left[ a b c \right]^T` (each vector corresponds to a row).

    The vectors a, b, c lie in the same plane if and only if the scalar
    triple product is 0.

    Parameters
    ----------
    a : array, shape (3,)
        Vector a.

    b : array shape (3,)
        Vector b.

    c : array, shape (3,)
        Vector c.

    Returns
    -------
    d : float
        Scalar triple product.
    """
    return np.dot(a, np.cross(b, c))


@numba.njit(numba.types.Tuple(
    (numba.float64[::1], numba.float64[::1]))(numba.float64[::1]), cache=True)
def plane_basis_from_normal(plane_normal):
    """Compute two basis vectors of a plane from the plane's normal vector.

    Note that there are infinitely many solutions because any rotation of the
    basis vectors about the normal is also a solution. This function
    deterministically picks one of the solutions.

    The two basis vectors of the plane together with the normal form an
    orthonormal basis in 3D space and could be used as columns to form a
    rotation matrix.

    Parameters
    ----------
    plane_normal : array-like, shape (3,)
        Plane normal of unit length.

    Returns
    -------
    x_axis : array, shape (3,)
        x-axis of the plane.

    y_axis : array, shape (3,)
        y-axis of the plane.
    """
    if abs(plane_normal[0]) >= abs(plane_normal[1]):
        # x or z is the largest magnitude component, swap them
        length = math.sqrt(
            plane_normal[0] * plane_normal[0]
            + plane_normal[2] * plane_normal[2])
        x_axis = np.array([-plane_normal[2] / length, 0.0,
                           plane_normal[0] / length])
        y_axis = np.array([
            plane_normal[1] * x_axis[2],
            plane_normal[2] * x_axis[0] - plane_normal[0] * x_axis[2],
            -plane_normal[1] * x_axis[0]])
    else:
        # y or z is the largest magnitude component, swap them
        length = math.sqrt(plane_normal[1] * plane_normal[1]
                           + plane_normal[2] * plane_normal[2])
        x_axis = np.array([0.0, plane_normal[2] / length,
                           -plane_normal[1] / length])
        y_axis = np.array([
            plane_normal[1] * x_axis[2] - plane_normal[2] * x_axis[1],
            -plane_normal[0] * x_axis[2], plane_normal[0] * x_axis[1]])
    return x_axis, y_axis


@numba.njit(numba.float64[::1](numba.float64[:, ::1], numba.float64[::1]),
            cache=True)
def transform_point(A2B, point_in_A):
    """Transform a point from frame A to frame B.

    Parameters
    ----------
    A2B : array, shape (4, 4)
        Transform from frame A to frame B as homogeneous matrix.

    point_in_A : array, shape (3,)
        Point in frame A.

    Returns
    -------
    point_in_B : array, shape (3,)
        Point in frame B.
    """
    return A2B[:3, 3] + np.dot(A2B[:3, :3], point_in_A)


@numba.njit(numba.float64[:, :](numba.float64[:, ::1], numba.float64[:, ::1]),
            cache=True)
def transform_points(A2B, points_in_A):
    """Transform points from frame A to frame B.

    Parameters
    ----------
    A2B : array, shape (4, 4)
        Transform from frame A to frame B as homogeneous matrix.

    points_in_A : array, shape (n_points, 3)
        Points in frame A.

    Returns
    -------
    points_in_A : array, shape (n_points, 3)
        Points in frame B.
    """
    return np.dot(points_in_A, A2B[:3, :3].T) + A2B[:3, 3]


@numba.njit(numba.float64[:, :](numba.float64[:, ::1], numba.float64[:, :]),
            cache=True)
def transform_directions(A2B, directions_in_A):
    """Transform directions from frame A to frame B.

    Parameters
    ----------
    A2B : array, shape (4, 4)
        Transform from frame A to frame B as homogeneous matrix.

    directions_in_A : array, shape (n_points, 3)
        Directions in frame A.

    Returns
    -------
    points_in_A : array, shape (n_points, 3)
        Points in frame B.
    """
    return np.dot(directions_in_A, A2B[:3, :3].T)


@numba.njit(numba.float64[::1](numba.float64[:, ::1], numba.float64[::1]),
            cache=True)
def inverse_transform_point(A2B, point_in_B):
    """Transform a point from frame B to frame A.

    Parameters
    ----------
    A2B : array, shape (4, 4)
        Transform from frame A to frame B as homogeneous matrix.

    point_in_B : array, shape (3,)
        Point in frame B.

    Returns
    -------
    point_in_A : array, shape (3,)
        Point in frame A.
    """
    RT = A2B[:3, :3].T
    return np.dot(RT, point_in_B) - np.dot(RT, A2B[:3, 3])


@numba.njit(numba.float64[:, :](numba.float64[:, :]), cache=True)
def invert_transform(A2B):
    """Invert transform.

    Parameters
    ----------
    A2B : array-like, shape (4, 4)
        Transform from frame A to frame B

    Returns
    -------
    B2A : array-like, shape (4, 4)
        Transform from frame B to frame A
    """
    B2A = np.empty((4, 4))
    RT = A2B[:3, :3].T
    B2A[:3, :3] = RT
    B2A[:3, 3] = -np.dot(RT, A2B[:3, 3])
    B2A[3, :3] = 0.0
    B2A[3, 3] = 1.0
    return B2A


@numba.njit(cache=True)
def cross_product_matrix(v):
    r"""Generate the cross-product matrix of a vector.

    The cross-product matrix :math:`\boldsymbol{V}` satisfies the equation

    .. math::

        \boldsymbol{V} \boldsymbol{w} = \boldsymbol{v} \times
        \boldsymbol{w}

    It is a skew-symmetric (antisymmetric) matrix, i.e.
    :math:`-\boldsymbol{V} = \boldsymbol{V}^T`.

    Parameters
    ----------
    v : array-like, shape (3,)
        3d vector

    Returns
    -------
    V : array-like, shape (3, 3)
        Cross-product matrix
    """
    return np.array([[0.0, -v[2], v[1]],
                     [v[2], 0.0, -v[0]],
                     [-v[1], v[0], 0.0]])


@numba.njit(cache=True)
def adjoint_from_transform(A2B):
    """Compute adjoint representation of a transformation matrix.

    The adjoint representation of a transformation
    :math:`\\left[Ad_{\\boldsymbol{T}_{BA}}\\right]`
    from frame A to frame B translates a twist from frame A to frame B
    through the adjoint map

    .. math::

        \\mathcal{V}_{B}
        = \\left[Ad_{\\boldsymbol{T}_{BA}}\\right] \\mathcal{V}_A

    The corresponding matrix form is

    .. math::

        \\left[\\mathcal{V}_{B}\\right]
        = \\boldsymbol{T}_{BA} \\left[\\mathcal{V}_A\\right]
        \\boldsymbol{T}_{BA}^{-1}

    We can also use the adjoint representation to transform a wrench from frame
    A to frame B:

    .. math::

        \\mathcal{F}_B
        = \\left[ Ad_{\\boldsymbol{T}_{AB}} \\right]^T \\mathcal{F}_A

    Note that not only the adjoint is transposed but also the transformation is
    inverted.

    Adjoint representations have the following properties:

    .. math::

        \\left[Ad_{\\boldsymbol{T}_1 \\boldsymbol{T}_2}\\right]
        = \\left[Ad_{\\boldsymbol{T}_1}\\right]
        \\left[Ad_{\\boldsymbol{T}_2}\\right]

    .. math::

        \\left[Ad_{\\boldsymbol{T}}\\right]^{-1} =
        \\left[Ad_{\\boldsymbol{T}^{-1}}\\right]

    Parameters
    ----------
    A2B : array-like, shape (4, 4)
        Transform from frame A to frame B

    Returns
    -------
    adj_A2B : array, shape (6, 6)
        Adjoint representation of transformation matrix
    """
    R = A2B[:3, :3]
    p = A2B[:3, 3]

    adj_A2B = np.zeros((6, 6))
    adj_A2B[:3, :3] = R
    adj_A2B[3:, :3] = np.dot(cross_product_matrix(p), R)
    adj_A2B[3:, 3:] = R
    return adj_A2B


def angles_between_vectors(A, B):
    """Compute angle between two vectors.

    Parameters
    ----------
    A : array, shape (n_vectors, 3)
        3d vectors

    B : array-like, shape (n_vectors, 3)
        3d vectors

    Returns
    -------
    angles : array, shape (n_vectors,)
        Angles between pairs of vectors from A and B
    """
    return np.arccos(
        np.clip(np.sum(A * B, axis=1) / (np.linalg.norm(A, axis=1)
                                         * np.linalg.norm(B, axis=1)),
                -1.0, 1.0))



"""Features related to input and output from and to disk."""
import os
import numpy as np


def load_mesh(filename, scale=1.0):
    """Load mesh from file.

    This functions attempts to load the mesh with Open3D first. If this fails,
    either because the library is not available or because the mesh format
    is not supported, it tries to use trimesh to load the file.

    Parameters
    ----------
    filename : str
        Path to mesh file.

    scale : float, optional (default: 1)
        Scale of vertex coordinates.

    Returns
    -------
    vertices : array, shape (n_vertices, 3)
        Vertices of the mesh.

    triangles : array, shape (n_triangles, 3)
        Indices of vertices that form triangles of the mesh.
    """
    try:
        import open3d as o3d
        mesh = o3d.io.read_triangle_mesh(filename)
        vertices = np.asarray(mesh.vertices)
        triangles = np.asarray(mesh.triangles)
        try_trimesh = False
    except OSError:
        try_trimesh = True
    except ImportError:
        try_trimesh = True

    if try_trimesh:  # pragma: no cover
        import trimesh
        mesh = trimesh.load(filename)
        vertices = np.asarray(mesh.vertices)
        triangles = np.asarray(mesh.faces)

    vertices *= scale
    return vertices, triangles


def load_tetrahedral_mesh(filename, scale=1.0):
    """Load tetrahedral mesh from file.

    Tetrahedral meshes are used mainly for simulation of deformable objects.

    Note that only the VTK format is currently supported
    (https://vtk.org/wp-content/uploads/2015/04/file-formats.pdf). You can
    use TetWild (https://github.com/Yixin-Hu/TetWild) to convert triangular
    meshes to tetrahedral meshes.

    Parameters
    ----------
    filename : str
        Path to mesh file.

    scale : float, optional (default: 1)
        Scale of vertex coordinates.

    Returns
    -------
    vertices : array, shape (n_vertices, 3)
        Vertices of the mesh.

    tetrahedra : array, shape (n_tetrahedra, 4)
        Indices of vertices that form tetrahedra of the mesh.
    """
    assert filename.endswith(".vtk")

    with open(filename, "r") as f:
        content = f.read()
    lines = content.splitlines()
    lines = [line.strip() for line in lines]

    point_lines = None
    cells_lines = None
    for i, line in enumerate(lines):
        if line.startswith("POINTS"):
            n_points = int(line.split(" ")[1])
            points_start = i + 1
            points_end = points_start + n_points
            point_lines = lines[points_start:points_end]
        elif line.startswith("CELLS"):
            n_cells = int(line.split(" ")[1])
            cells_start = i + 1
            cells_end = cells_start + n_cells
            cells_lines = lines[cells_start:cells_end]
    assert point_lines is not None
    assert cells_lines is not None

    points = np.row_stack([np.fromstring(line, sep=" ", dtype=float)
                           for line in point_lines])
    cells = np.row_stack([np.fromstring(line, sep=" ", dtype=int)
                          for line in cells_lines])
    assert all(cells[:, 0] == 4)

    vertices = points * scale
    tetrahedra = cells[:, 1:]

    return vertices, tetrahedra



import numpy as np


def barycentric_transforms(tetrahedra_points):
    """Returns X. X.dot(coords) = (r, 1), where r is a Cartesian vector."""
    # NOTE that in the original paper it is not obvious that we have to take
    # the inverse
    return np.linalg.pinv(np.hstack((tetrahedra_points.transpose((0, 2, 1)),
                                     np.ones((len(tetrahedra_points), 1, 4)))))



"""Broad-phase collision detection."""
import warnings

import numpy as np
from pytransform3d import urdf

from .aabb_tree import AabbTree
from .colliders import Sphere, Box, Cylinder, MeshGraph
from .urdf_utils import self_collision_whitelists
from .io import load_mesh


class BoundingVolumeHierarchy:
    """Bounding volume hierarchy (BVH) for broad phase collision detection.

    Wraps multiple colliders that are connected through transformations.
    In addition, these colliders are stored in an AABB tree for broad phase
    collision detection.

    Parameters
    ----------
    tm : pytransform3d.transform_manager.TransformManager
        Transform manager that stores the transformations.

    base_frame : str
        Name of the base frame in which colliders are represented.

    base_frame2origin : array, shape (4, 4), optional (default: np.eye(4))
        The position of the base_frame to origin.

    Attributes
    ----------
    aabbtree_ : AabbTree
        Tree of axis-aligned bounding boxes.

    colliders_ : dict
        Maps frames of collision objects to colliders.

    self_collision_whitelists_ : dict
        Whitelists for self-collision detection in case this BVH represents
        a robot.
    """
    def __init__(self, tm, base_frame, base_frame2origin=np.eye(4)):
        tm.add_transform(base_frame, "origin", base_frame2origin)
        self.tm = tm
        self.base_frame = base_frame
        self.base_frame2origin = base_frame2origin
        self.collider_frames = set()
        self.aabbtree_ = AabbTree()
        self.colliders_ = {}
        self.self_collision_whitelists_ = {}

    def fill_tree_with_colliders(
            self, tm, make_artists=False,
            fill_self_collision_whitelists=False, use_visuals=False):
        """Fill tree with colliders from URDF transform manager.

        Parameters
        ----------
        tm : pytransform3d.urdf.UrdfTransformManager
            Transform manager that has colliders.

        make_artists : bool, optional (default: False)
            Create artist for visualization for each collision object.

        fill_self_collision_whitelists : bool, optional (default: False)
            Fill whitelists for self collision detection. All collision
            objects connected to the current link, child, and parent links
            will be ignored.

        use_visuals : bool, optional (default: False)
            Use visual objects as colliders.
        """
        if use_visuals:
            objects = tm.visuals
        else:
            objects = tm.collision_objects

        for obj in objects:
            try:
                collider = self._make_collider(tm, obj, make_artists)
                self.add_collider(obj.frame, collider)
            except RuntimeError as e:
                warnings.warn(str(e))

        if fill_self_collision_whitelists:
            self.self_collision_whitelists_.update(
                self_collision_whitelists(tm))

        self.update_collider_poses()

    def _make_collider(self, tm, obj, make_artists):
        """Creates a collider from a URDF object.

        Parameters
        ----------
        tm : pytransform3d.urdf.UrdfTransformManager
            Transform manager that has colliders.

        obj: pytransform3d.urdf.Geometry
            The urdf object.

        make_artists : bool, optional (default: False)
            Create artist for visualization for each collision object.

        Returns
        -------
        collider : ConvexCollider
            The corresponding collider.
        """
        A2B = tm.get_transform(obj.frame, "origin")

        if isinstance(obj, urdf.Sphere):
            collider = Sphere(center=A2B[:3, 3], radius=obj.radius)
        elif isinstance(obj, urdf.Box):
            collider = Box(A2B, obj.size)
        elif isinstance(obj, urdf.Cylinder):
            collider = Cylinder(
                cylinder2origin=A2B, radius=obj.radius,
                length=obj.length)
        else:
            assert isinstance(obj, urdf.Mesh)
            vertices, triangles = load_mesh(obj.filename, obj.scale)
            collider = MeshGraph(A2B, vertices, triangles)
        if make_artists:
            collider.make_artist()
        return collider

    def add_collider(self, frame, collider):
        """Add collider.

        Parameters
        ----------
        frame : Hashable
            Frame in which the collider is located.

        collider : ConvexCollider
            Collider.
        """
        self.collider_frames.add(frame)
        self.colliders_[frame] = collider
        self.aabbtree_.insert_aabb(collider.aabb(), (frame, collider))

    def update_collider_poses(self):
        """Update poses of all colliders from transform manager."""
        self.aabbtree_ = AabbTree()
        for frame in self.colliders_:
            A2B = self.tm.get_transform(frame, "origin")
            collider = self.colliders_[frame]
            collider.update_pose(A2B)
            self.aabbtree_.insert_aabb(collider.aabb(), (frame, collider))

    def get_colliders(self):
        """Get all colliders.

        Returns
        -------
        colliders : list
            List of colliders.
        """
        return self.colliders_.values()

    def get_artists(self):
        """Get all artists.

        Returns
        -------
        artists : list
            List of artists.
        """
        return [collider.artist_ for collider in self.colliders_.values()
                if collider.artist_ is not None]

    def aabb_overlapping_colliders(self, collider, whitelist=()):
        """Get colliders with an overlapping AABB.

        This function performs broad phase collision detection with a bounding
        volume hierarchy, where the bounding volumes are axis-aligned bounding
        boxes.

        Parameters
        ----------
        collider : ConvexCollider
            Collider.

        whitelist : sequence
            Names of frames to which collisions are allowed.

        Returns
        -------
        colliders : dict
            Maps frame names to colliders with overlapping AABB.
        """
        aabb = collider.aabb()
        _, overlaps = self.aabbtree_.overlaps_aabb(aabb)
        colliders = dict(np.array(self.aabbtree_.external_data_list,
                                  dtype=object)[overlaps.astype(int)])
        for frame in whitelist:
            colliders.pop(frame, None)
        return colliders

    def aabb_overlapping_with_other_bvh(self, other_bvh):
        """Get colliders with an overlapping AABB.

        This function performs broad phase collision detection with a bounding
        volume hierarchy, where the bounding volumes are axis-aligned bounding
        boxes.

        Parameters
        ----------
        other_bvh : BoundingVolumeHierarchy
            the other BVH.

        Returns
        -------
        data_pairs : array, shape(n, 2)
            A list of colliding colliders.
        """
        _, _, _, pairs = self.aabbtree_.overlaps_aabb_tree(other_bvh.aabbtree_)

        data_pairs = []
        for pair in pairs:
            data_pair = (self.aabbtree_.external_data_list[pair[0]],
                         other_bvh.aabbtree_.external_data_list[pair[1]])
            data_pairs.append(data_pair)

        return data_pairs

    def aabb_overlapping_with_self(self):
        """Get colliders with overlapping with itself.

        This function performs broad phase collision detection with a bounding
        volume hierarchy, where the bounding volumes are axis-aligned bounding
        boxes.

        Returns
        -------
        data_pairs : array, shape(n, 2)
            A list of colliding colliders.
        """
        _, _, _, pairs = self.aabbtree_.overlaps_aabb_tree(self.aabbtree_)

        data_pairs = []
        for pair in pairs:
            if pair[0] == pair[1]:
                continue

            data_pair = (self.aabbtree_.external_data_list[pair[0]],
                         self.aabbtree_.external_data_list[pair[1]])
            data_pairs.append(data_pair)

        return data_pairs

    def get_collider_frames(self):
        """Get collider frames.

        Returns
        -------
        collider_frames : set
            Collider frames.
        """
        return self.collider_frames



import numba
import numpy as np
from ..utils import plane_basis_from_normal, EPSILON
from ._halfplanes import intersect_halfplanes


def intersect_tetrahedron_pairs(
        pairs, tetrahedra_points1, tetrahedra_points2,
        epsilon1, epsilon2, X1, X2, youngs_modulus1=1.0, youngs_modulus2=1.0):
    """Intersect pairs of tetrahedra.

    Parameters
    ----------
    pairs : list
        List of index pairs.

    tetrahedra_points1 : array, shape (n_tetrahedra1, 4, 3)
        Points that form the tetrahedra of object 1.

    tetrahedra_points2 : array, shape (n_tetrahedra2, 4, 3)
        Points that form the tetrahedra of object 2.

    epsilon1 : array, shape (n_tetrahedra1, 4)
        Potentials of the vertices of the tetrahedra of object 1.

    epsilon2 : array, shape (n_tetrahedra2, 4)
        Potentials of the vertices of the tetrahedra of object 2.

    X1 : dict
        Maps tetrahedron indices of first rigid body to barycentric transform.

    X2 : dict
        Maps tetrahedron indices of second rigid body to barycentric transform.

    youngs_modulus1 : float, optional (default: 1.0)
        Young's modulus of the first rigid body

    youngs_modulus2 : float, optional (default: 1.0)
        Young's modulus of the second rigid body

    Returns
    -------
    intersection : bool
        Do both rigid bodies overlap?

    contact_planes : array, shape (n_intersections, 4)
        Contact planes of intersection pairs in Hesse normal form.

    contact_polygons : list
        Vertices of contact polygons in counter-clockwise order.

    intersecting_tetrahedra1 : list
        Intersecting tetrahedron indices of first mesh.

    intersecting_tetrahedra2 : list
        Intersecting tetrahedron indices of second mesh.
    """
    intersection = False
    contact_planes = []
    contact_polygons = []
    intersecting_tetrahedra1 = []
    intersecting_tetrahedra2 = []
    for i, j in pairs:
        intersecting, contact_details = intersect_tetrahedron_pair(
            tetrahedra_points1[i], epsilon1[i], X1[i],
            tetrahedra_points2[j], epsilon2[j], X2[j],
            youngs_modulus1, youngs_modulus2)
        if intersecting:
            intersection = True
        else:
            continue

        contact_plane_hnf, contact_polygon = contact_details

        intersecting_tetrahedra1.append(i)
        intersecting_tetrahedra2.append(j)
        contact_planes.append(contact_plane_hnf)
        contact_polygons.append(contact_polygon)
    if len(contact_planes) > 0:
        contact_planes = np.vstack(contact_planes)
    return (intersection, contact_planes, contact_polygons,
            intersecting_tetrahedra1, intersecting_tetrahedra2)


@numba.njit(cache=True)
def intersect_tetrahedron_pair(tetrahedron1, epsilon1, X1,
                               tetrahedron2, epsilon2, X2,
                               youngs_modulus1=1.0, youngs_modulus2=1.0):
    """Intersect a pair of tetrahedra.

    Parameters
    ----------
    tetrahedron1 : array, shape (4, 3)
        Vertices of tetrahedron 1.

    epsilon1 : array, shape (4,)
        Potentials of the vertices of tetrahedron 1.

    X1 : array, shape (4, 4)
        Each row is a halfspace that defines the original tetrahedron.

    tetrahedron2 : array, shape (4, 3)
        Vertices of tetrahedron 2.

    epsilon2 : array, shape (4,)
        Potentials of the vertices of tetrahedron 2.

    X2 : array, shape (4, 4)
        Each row is a halfspace that defines the original tetrahedron.

    youngs_modulus1 : float, optional (default: 1.0)
        Young's modulus of tetrahedron 1

    youngs_modulus2 : float, optional (default: 1.0)
        Young's modulus of tetrahedron 2

    Returns
    -------
    intersection : bool
        Do both tetrahedra intersect?

    contact_info : tuple
        Contact plane in Hesse normal form, contact polygon and triangles of
        contact polygon.
    """
    contact_plane_hnf, same = contact_plane(X1, X2, epsilon1, epsilon2, youngs_modulus1, youngs_modulus2)
    if same:
        return True, _handle_same_tetrahedron(epsilon2, tetrahedron2)

    plane_normal = contact_plane_hnf[:3]
    d = contact_plane_hnf[3]
    if not check_tetrahedra_intersect_contact_plane(
            tetrahedron1, tetrahedron2, plane_normal, d, tolerance=1e-6):
        return False, (contact_plane_hnf, None)

    contact_polygon = compute_contact_polygon(X1, X2, plane_normal, d)
    if len(contact_polygon) < 3:
        return False, (contact_plane_hnf, contact_polygon)

    return True, (contact_plane_hnf, contact_polygon)


@numba.njit(cache=True)
def _handle_same_tetrahedron(epsilon, tetrahedron):
    """Generate contact plane and polygon for edge case of same tetrahedra.

    In this case the contact plane passes through the potential-weighted center
    of the tetrahedron.
    """
    weights = epsilon / sum(epsilon)
    plane_point = weights.dot(tetrahedron)
    d = np.linalg.norm(plane_point)
    if d > 0.0:
        plane_normal = plane_point / d
    else:  # any direction is possible
        plane_normal = np.array([0.0, 0.0, 1.0], dtype=np.dtype("float"))
    contact_plane_hnf = np.array([
        plane_normal[0], plane_normal[1], plane_normal[2], d])
    contact_polygon = np.vstack((plane_point, plane_point, plane_point))
    return contact_plane_hnf, contact_polygon


@numba.njit(
    numba.types.Tuple((numba.float64[::1], numba.bool_))(
        numba.float64[:, ::1], numba.float64[:, ::1],
        numba.float64[::1], numba.float64[::1],
        numba.float64, numba.float64),
    cache=True)
def contact_plane(X1, X2, epsilon1, epsilon2, youngs_modulus1, youngs_modulus2):
    """Compute contact plane.

    Parameters
    ----------
    X1 : array, shape (4, 4)
        Each row is a halfspace that defines the original tetrahedron.

    X2 : array, shape (4, 4)
        Each row is a halfspace that defines the original tetrahedron.

    epsilon1 : array, shape (4,)
        Potentials of the vertices of tetrahedron 1.

    epsilon2 : array, shape (4,)
        Potentials of the vertices of tetrahedron 2.

    youngs_modulus1 : float, optional (default: 1.0)
        Young's modulus of tetrahedron 1

    youngs_modulus2 : float, optional (default: 1.0)
        Young's modulus of tetrahedron 2

    Returns
    -------
    plane_hnf : array, shape (4,)
        Contact plane in Hesse normal form: plane normal and distance to origin
        along plane normal.

    same : bool
        Are both tetrahedrons actually the same tetrahedron (maybe scaled)?
    """
    plane_hnf = (epsilon1 * youngs_modulus1).dot(X1) - (epsilon2 * youngs_modulus2).dot(X2)
    norm = np.linalg.norm(plane_hnf[:3])

    if norm == 0.0:
        return plane_hnf, True

    plane_hnf /= norm
    # NOTE in order to obtain proper Hesse normal form of the contact plane
    # we have to multiply the scalar by -1, since it appears as -d in the
    # equation np.dot(normal, x) - d = 0. However, it appears as
    # np.dot(normal, x) + d = 0 in the paper (page 7).
    plane_hnf[3] *= -1

    if abs(plane_hnf[3]) < 10.0 * EPSILON:
        return plane_hnf, True

    return plane_hnf, False


@numba.njit(
    numba.bool_(numba.float64[:, ::1], numba.float64[:, ::1],
                numba.float64[::1], numba.float64, numba.float64),
    cache=True)
def check_tetrahedra_intersect_contact_plane(
        tetrahedron1, tetrahedron2, plane_normal, d, tolerance):
    """Check if tetrahedra intersect contact plane.

    Parameters
    ----------
    tetrahedron1 : array, shape (4, 3)
        Vertices of tetrahedron 1.

    tetrahedron2 : array, shape (4, 3)
        Vertices of tetrahedron 2.

    plane_normal : array, shape (3,)
        Normal of the contact plane.

    d : float
        Distance to origin along normal.

    tolerance : float, optional (default: 1e-6)
        Floating point tolerance.

    Returns
    -------
    intersection : bool
        Do both tetrahedra intersect the contact plane.
    """
    plane_distances1 = tetrahedron1.dot(plane_normal) - d
    plane_distances2 = tetrahedron2.dot(plane_normal) - d
    return (
            min(plane_distances1) < -tolerance
            and max(plane_distances1) > tolerance
            and min(plane_distances2) < -tolerance
            and max(plane_distances2) > tolerance)


@numba.njit(
    numba.float64[:, ::1](numba.float64[:, ::1], numba.float64[::1],
                          numba.float64[:, ::1]),
    cache=True)
def make_halfplanes(X, plane_point, cart2plane):
    """Project triangles of a tetrahedron to contact plane.

    Parameters
    ----------
    X : array, shape (8, 4)
        Each row is a halfspace that defines one of the original tetrahedra.

    plane_point : array, shape (3,)
        Point on plane.

    cart2plane : array, shape (2, 3)
        Projection from 3D Cartesian space to contact plane.

    Returns
    -------
    halfplanes : array, shape (n_halfplanes, 4)
        Halfplanes in contact plane. Each halfplane is defined by a point
        p and a direction pq.
    """
    halfplanes = np.empty((8, 4))
    face_normals = np.ascontiguousarray(X[:, :3])
    normals2d = face_normals.dot(cart2plane.T)
    ds = -X[:, 3] - face_normals.dot(plane_point)
    hp_idx = 0
    for i in range(8):
        norm = np.linalg.norm(normals2d[i])
        if norm > EPSILON:
            p = normals2d[i] * ds[i] / (norm * norm)
            halfplanes[i, :2] = p
            halfplanes[i, 2] = normals2d[i, 1]
            halfplanes[i, 3] = -normals2d[i, 0]
            hp_idx += 1
    return halfplanes[:hp_idx]


@numba.njit(
    numba.float64[:, ::1](numba.float64[:, ::1]),
    cache=True)
def order_points(points):
    """Order points by angle around their center.

    Parameters
    ----------
    points : array, shape (n_points, 2)
        Points.

    Returns
    -------
    ordered_points : shape (n_points, 2)
        Ordered points.
    """
    center = np.array([np.mean(points[:, 0]), np.mean(points[:, 1])])
    poly_centered = points - center
    angles = np.arctan2(poly_centered[:, 1], poly_centered[:, 0])
    return points[np.argsort(angles)]


@numba.njit(
    numba.float64[:, ::1](numba.float64[:, ::1]),
    cache=True)
def filter_unique_points(points):
    """Remove duplicate points.

    Parameters
    ----------
    points :  array, shape (n_points, 2)
        Points that should be filtered.

    Returns
    -------
    unique_points : array, shape (n_unique_points, 2)
        Unique points.
    """
    epsilon = 10.0 * EPSILON
    unique_points = np.empty((len(points), 2))
    n_unique_points = 0
    for j in range(len(points)):
        if j == 0 or np.linalg.norm(points[j] - points[j - 1]) > epsilon:
            unique_points[n_unique_points] = points[j]
            n_unique_points += 1
    return unique_points[:n_unique_points]


@numba.njit(
    numba.float64[:, ::1](numba.float64[:, ::1], numba.float64[:, ::1],
                          numba.float64[::1]),
    cache=True)
def project_polygon_to_3d(vertices, cart2plane, plane_point):
    """Project polygon from contact plane to 3D space.

    Parameters
    ----------
    vertices : array, shape (n_vertices, 2)
        Vertices of contact polygon in contact plane.

    cart2plane : array, shape (2, 3)
        Projection from 3D Cartesian space to contact plane.

    plane_point : array, shape (3,)
        Point on plane. Projection offset from contact plane to 3D space.

    Returns
    -------
    vertices3d : array, shape (n_vertices, 3)
        Vertices of contact polygon in 3D space.
    """
    # Note that we actually apply the transform plane2cart to the vertices,
    # however, it is the transpose of cart2plane and to apply it to a batch
    # of vertices, we also have to transpose it, so both transposes cancel out.
    return vertices.dot(cart2plane) + plane_point


@numba.njit(
    numba.float64[:, :](numba.float64[:, ::1], numba.float64[:, ::1],
                        numba.float64[::1], numba.float64),
    cache=True)
def compute_contact_polygon(X1, X2, plane_normal, d):
    """Compute contact polygon.

    Parameters
    ----------
    X1 : array, shape (4, 4)
        Each row is a halfspace that defines the original tetrahedron.

    X2 : array, shape (4, 4)
        Each row is a halfspace that defines the original tetrahedron.

    plane_normal : array, shape (3,)
        Normal of the contact plane.

    d : float
        Distance to origin along normal.

    Returns
    -------
    polygon3d : array, shape (n_vertices, 3)
        Contact polygon between two tetrahedra. Points are ordered
        counter-clockwise around their center. No intersection is indicated
        by 0 vertices.
    """
    plane_point = plane_normal * d
    cart2plane = np.vstack(plane_basis_from_normal(plane_normal))
    halfplanes = make_halfplanes(np.vstack((X1, X2)), plane_point, cart2plane)

    vertices2d = intersect_halfplanes(halfplanes)
    if len(vertices2d) < 3:
        return np.empty((0, 3), dtype=np.dtype("float"))

    # this approach sometimes results in duplicate points, remove them
    vertices2d = order_points(np.ascontiguousarray(vertices2d))
    unique_vertices2d = filter_unique_points(vertices2d)
    if len(unique_vertices2d) < 3:
        return np.empty((0, 3), dtype=np.dtype("float"))

    # from ._halfplanes import plot_halfplanes_and_intersections
    # plot_halfplanes_and_intersections(halfplanes, unique_vertices2d)

    vertices3d = project_polygon_to_3d(unique_vertices2d, cart2plane,
                                       plane_point)
    return vertices3d



import numpy as np
from .utils import EPSILON, invert_transform


def points_in_sphere(points, center, radius):
    """Test if points are in sphere.

    Parameters
    ----------
    points : array, shape (n_points, 3)
        Points.

    center : array, shape (3,)
        Center of the sphere.

    radius : float
        Radius of the sphere.

    Returns
    -------
    contained : array, shape (n_points,)
        Boolean array indicating whether each point is contained or not.
    """
    diff = points - center
    squared_dist = np.sum(diff * diff, axis=1)
    return squared_dist <= radius * radius


def points_in_capsule(points, capsule2origin, radius, height):
    """Test if points are in capsule.

    Parameters
    ----------
    points : array, shape (n_points, 3)
        Points.

    capsule2origin : array, shape (4, 4)
        Pose of the capsule.

    radius : float
        Radius of the capsule.

    height : float
        Height of the capsule.

    Returns
    -------
    contained : array, shape (n_points,)
        Boolean array indicating whether each point is contained or not.
    """
    segment_start = capsule2origin[:3, 3] - 0.5 * height * capsule2origin[:3, 2]
    segment_end = capsule2origin[:3, 3] + 0.5 * height * capsule2origin[:3, 2]
    segment_direction = segment_end - segment_start
    t = (np.dot(points - segment_start, segment_direction) /
         np.dot(segment_direction, segment_direction))
    t = np.minimum(np.maximum(t, 0.0), 1.0)
    closest_points_line_segment = (
        segment_start[np.newaxis] +
        t[:, np.newaxis] * segment_direction[np.newaxis])
    diff = points - closest_points_line_segment
    squared_dist = np.sum(diff * diff, axis=1)
    return squared_dist <= radius * radius


def points_in_ellipsoid(points, ellipsoid2origin, radii):
    """Test if points are in ellipsoid.

    Parameters
    ----------
    points : array, shape (n_points, 3)
        Points.

    ellipsoid2origin : array, shape (4, 4)
        Pose of the ellipsoid.

    radii : array, shape (3,)
        Radii of the ellipsoid.

    Returns
    -------
    contained : array, shape (n_points,)
        Boolean array indicating whether each point is contained or not.
    """
    origin2ellipsoid = invert_transform(ellipsoid2origin)
    points = origin2ellipsoid[:3, 3] + np.dot(points, origin2ellipsoid[:3, :3].T)
    normalized_points = points / radii
    return np.sum(normalized_points * normalized_points, axis=1) <= 1.0


def points_in_disk(points, center, radius, normal):
    """Test if points are in disk.

    Parameters
    ----------
    points : array, shape (n_points, 3)
        Points.

    center : array, shape (3,)
        Center of the disk.

    radius : float
        Radius of the disk.

    normal : array, shape (3,)
        Normal to the plane in which the disk lies.

    Returns
    -------
    contained : array, shape (n_points,)
        Boolean array indicating whether each point is contained or not.
    """
    contained = np.empty(len(points), dtype=bool)
    contained[:] = True
    # signed distance from point to plane of disk
    diff = points - center
    dist_to_plane = diff.dot(normal)
    contained[np.abs(dist_to_plane) > 10.0 * EPSILON] = False

    # projection of P - C onto plane is Q - C = P - C - dist_to_plane * N
    diff_in_plane = diff - dist_to_plane[:, np.newaxis] * normal[np.newaxis]
    sqr_dist_in_plane = np.sum(diff_in_plane * diff_in_plane, axis=1)
    contained[sqr_dist_in_plane > radius * radius] = False
    return contained


def points_in_cone(points, cone2origin, radius, height):
    """Test if points are in cone.

    Parameters
    ----------
    points : array, shape (n_points, 3)
        Points.

    cone2origin : array, shape (4, 4)
        Pose of the cone.

    radius : float
        Radius of the cone.

    height : float
        Length of the cone.

    Returns
    -------
    contained : array, shape (n_points,)
        Boolean array indicating whether each point is contained or not.
    """
    contained = np.empty(len(points), dtype=bool)
    contained[:] = True
    half_height = 0.5 * height
    # signed distance from point to plane of cone center
    diff = points - (cone2origin[:3, 3] + half_height * cone2origin[:3, 2])
    dist_to_center_plane = diff.dot(cone2origin[:3, 2])
    outside_z = np.abs(dist_to_center_plane) > half_height
    inside_z = np.logical_not(outside_z)
    contained[outside_z] = False

    # projection of P - C onto plane is Q - C = P - C - dist_to_plane * N
    diff_in_plane = diff[inside_z] - dist_to_center_plane[inside_z, np.newaxis] * cone2origin[np.newaxis, :3, 2]
    sqr_dist_in_plane = np.sum(diff_in_plane * diff_in_plane, axis=1)
    dist_to_base_plane = dist_to_center_plane[inside_z] + half_height
    radii = (1.0 - dist_to_base_plane / height) * radius
    not_contained = sqr_dist_in_plane > radii * radii
    inside_z_indices = np.where(inside_z)[0]
    not_contained_indices = inside_z_indices[not_contained]
    contained[not_contained_indices] = False
    return contained


def points_in_cylinder(points, cylinder2origin, radius, length):
    """Test if points are in cylinder.

    Parameters
    ----------
    points : array, shape (n_points, 3)
        Points.

    cylinder2origin : array, shape (4, 4)
        Pose of the cylinder.

    radius : float
        Radius of the cylinder.

    length : float
        Length of the cylinder.

    Returns
    -------
    contained : array, shape (n_points,)
        Boolean array indicating whether each point is contained or not.
    """
    contained = np.empty(len(points), dtype=bool)
    contained[:] = True
    # signed distance from point to plane of disk
    diff = points - cylinder2origin[:3, 3]
    dist_to_plane = diff.dot(cylinder2origin[:3, 2])
    contained[np.abs(dist_to_plane) > 0.5 * length] = False

    # projection of P - C onto plane is Q - C = P - C - dist_to_plane * N
    diff_in_plane = diff - dist_to_plane[:, np.newaxis] * cylinder2origin[np.newaxis, :3, 2]
    sqr_dist_in_plane = np.sum(diff_in_plane * diff_in_plane, axis=1)
    contained[sqr_dist_in_plane > radius * radius] = False
    return contained


def points_in_box(points, box2origin, size):
    """Test if points are in box.

    Parameters
    ----------
    points : array, shape (n_points, 3)
        Points.

    box2origin : array, shape (4, 4)
        Pose of the box.

    size : array, shape (3,)
        Sizes of the box along its axes.

    Returns
    -------
    contained : array, shape (n_points,)
        Boolean array indicating whether each point is contained or not.
    """
    origin2box = invert_transform(box2origin)
    points = origin2box[:3, 3] + np.dot(points, origin2box[:3, :3].T)
    return np.all(np.abs(points) <= 0.5 * size, axis=1)


def points_in_convex_mesh(points, mesh2origin, vertices, triangles):
    """Test if points are in convex mesh.

    Parameters
    ----------
    points : array, shape (n_points, 3)
        Points.

    mesh2origin : array, shape (4, 4)
        Pose of the mesh.

    vertices : array, shape (n_vertices, 3)
        Vertices of the mesh.

    triangles : array, shape (n_triangles, 3)
        Indices of vertices that form triangles of the mesh. Face normals
        must point outwards.

    Returns
    -------
    contained : array, shape (n_points,)
        Boolean array indicating whether each point is contained or not.
    """
    origin2mesh = invert_transform(mesh2origin)
    points = origin2mesh[:3, 3] + np.dot(points, origin2mesh[:3, :3].T)

    faces = vertices[triangles]
    A = faces[:, 1] - faces[:, 0]
    B = faces[:, 2] - faces[:, 0]
    face_normals = np.cross(A, B)
    face_centers = np.mean(faces, axis=1)
    contained = np.empty(len(points), dtype=bool)
    contained[:] = True
    for i, point in enumerate(points):
        normal_projected_points = np.sum(
            face_normals * (point[np.newaxis] - face_centers), axis=1)
        if np.any(normal_projected_points > 0.0):
            contained[i] = False
    return contained



import numba
import numpy as np
from ..utils import adjoint_from_transform


def contact_surface_forces(contact_surface, rigid_body1):
    tetrahedra_potentials1 = rigid_body1.tetrahedra_potentials
    n_intersections = len(contact_surface.intersecting_tetrahedra1)
    coms = np.empty((n_intersections, 3), dtype=float)
    forces = np.empty((n_intersections, 3), dtype=float)
    areas = np.empty(n_intersections, dtype=float)
    triangles = []
    for intersection_idx, tetrahedron_idx in enumerate(
            contact_surface.intersecting_tetrahedra1):
        coms[intersection_idx], forces[intersection_idx], \
        areas[intersection_idx], triangle = compute_contact_force(
            rigid_body1.tetrahedra_points[tetrahedron_idx],
            tetrahedra_potentials1[tetrahedron_idx],
            contact_surface.contact_planes[intersection_idx],
            contact_surface.contact_polygons[intersection_idx],
            rigid_body1.youngs_modulus)

        triangles.append(triangle)
    return areas, coms, forces, triangles


@numba.njit(cache=True)
def tesselate_ordered_polygon(n_vertices):
    """Tesselate a ccw-ordered polygon.

    Parameters
    ----------
    n_vertices : int
        Number of vertices of the polygon.

    Returns
    -------
    triangles : array, shape (n_vertices - 2, 3)
        Triangles forming the polygon.
    """
    triangles = np.empty((n_vertices - 2, 3), dtype=np.dtype("int"))
    triangles[:, 0] = 0
    triangles[:, 1] = np.arange(1, n_vertices - 1)
    triangles[:, 2] = np.arange(2, n_vertices)
    return triangles


# 8 halfplanes cannot define a polygon with more than 8 vertices
TRIANGLES = tesselate_ordered_polygon(8)


@numba.njit(cache=True)
def compute_contact_force(
        tetrahedron, epsilon, contact_plane_hnf, contact_polygon, youngs_modulus=1.0):
    """Compute contact force from contact polygon.

    Parameters
    ----------
    tetrahedron : array, shape (4, 3)
        Vertices of tetrahedron

    epsilon : array, shape (4,)
        Potentials of vertices of tetrahedron.

    contact_plane_hnf : array, shape (4,)
        Contact plane in Hesse normal form.

    contact_polygon : array, shape (n_vertices, 3)
        Contact polygon between two tetrahedra. Points are ordered
        counter-clockwise around their center.

    youngs_modulus : float, optional (default: 1.0)
        The stiffness of the material.

    Returns
    -------
    intersection_com : array, shape (3,)
        Center of the contact polygon.

    force_vector : array, shape (3,)
        Force vector of this contact.

    total_area : float
        Area of contact polygon.

    triangles : array, shape (n_triangles, 3)
        Vertex indices of triangles.
    """
    total_force = 0.0
    total_area = 0.0
    intersection_com = np.zeros(3)

    X = np.vstack((tetrahedron.T, np.ones((1, 4))))
    com = np.empty(4, dtype=np.dtype("float"))
    com[3] = 1.0
    triangles = TRIANGLES[:len(contact_polygon) - 2]
    for triangle in triangles:
        vertices = contact_polygon[triangle]
        com[:3] = (vertices[0] + vertices[1] + vertices[2]) / 3.0
        res = np.linalg.solve(X, com)
        pressure = np.sum(res * (epsilon * youngs_modulus))
        area = 0.5 * np.linalg.norm(np.cross(vertices[1] - vertices[0],
                                             vertices[2] - vertices[0]))
        total_force += pressure * area
        total_area += area
        intersection_com += area * com[:3]

    if total_area > 0.0:
        intersection_com /= total_area
    else:
        intersection_com = contact_polygon[0]

    force_vector = total_force * contact_plane_hnf[:3]
    return intersection_com, force_vector, total_area, triangles


def accumulate_wrenches(contact_surface, rigid_body1, rigid_body2):
    total_force_21 = np.sum(contact_surface.contact_forces, axis=0)
    total_torque_21 = np.sum(
        np.cross(contact_surface.contact_coms - rigid_body1.com,
                 contact_surface.contact_forces), axis=0)
    total_torque_12 = np.sum(
        np.cross(contact_surface.contact_coms - rigid_body2.com,
                 -contact_surface.contact_forces), axis=0)
    wrench12_in_world, wrench21_in_world = _transform_wrenches(
        contact_surface.frame2world, total_force_21, total_torque_12,
        total_torque_21)
    return wrench12_in_world, wrench21_in_world


@numba.njit(cache=True)
def _transform_wrenches(
        mesh22origin, total_force_21, total_torque_12, total_torque_21):
    wrench21 = np.hstack((total_force_21, total_torque_21))
    wrench12 = np.hstack((-total_force_21, total_torque_12))
    mesh22origin_adjoint = adjoint_from_transform(mesh22origin)
    wrench21_in_world = mesh22origin_adjoint.T.dot(wrench21)
    wrench12_in_world = mesh22origin_adjoint.T.dot(wrench12)
    return wrench12_in_world, wrench21_in_world



import numba
import numpy as np
from ..utils import norm_vector, EPSILON


# replaces from numba.np.extensions import cross2d, which seems to have a bug
# when called with NUMBA_DISABLE_JIT=1
@numba.njit(
    numba.float64(numba.float64[::1], numba.float64[::1]),
    cache=True)
def cross2d(a, b):
    return a[0] * b[1] - a[1] * b[0]


@numba.njit(
    numba.float64[::1](numba.float64[::1], numba.float64[::1]),
    cache=True)
def intersect_two_halfplanes(halfplane1, halfplane2):
    denom = cross2d(halfplane1[2:], halfplane2[2:])
    if abs(denom) < EPSILON:
        return np.empty(0, dtype=np.dtype("float"))
    t = cross2d((halfplane2[:2] - halfplane1[:2]), halfplane2[2:]) / denom
    return halfplane1[:2] + halfplane1[2:] * t


@numba.njit(
    numba.bool_(numba.float64[::1], numba.float64[::1]),
    cache=True)
def point_outside_of_halfplane(halfplane, point):
    return cross2d(halfplane[2:], point - halfplane[:2]) < -EPSILON


@numba.njit(
    numba.float64[:, :](numba.float64[:, ::1]), cache=True)
def intersect_halfplanes(halfplanes):
    """Find polygon points by halfplane intersection.

    Parameters
    ----------
    halfplanes : array, shape (n_halfplanes, 4)
        Halfplanes in contact plane. Each halfplane is defined by a point
        p and a direction pq.

    Returns
    -------
    points : list
        Points of the polygon.
    """
    # reserve more space than required, there might be duplicates
    points = np.empty((3 * len(halfplanes), 2))
    n_intersections = 0
    for i in range(len(halfplanes)):
        for j in range(i + 1, len(halfplanes)):
            p = intersect_two_halfplanes(halfplanes[i], halfplanes[j])
            if len(p) == 0:  # parallel halfplanes
                continue
            valid = True
            for k in range(len(halfplanes)):
                if k != i and k != j and point_outside_of_halfplane(
                        halfplanes[k], p):
                    valid = False
                    break
            if valid:
                points[n_intersections] = p
                n_intersections += 1
    assert n_intersections < len(points)
    return points[:n_intersections]


def plot_halfplanes_and_intersections(halfplanes, points=None, xlim=None, ylim=None):  # pragma: no cover
    import matplotlib.pyplot as plt
    if points is None:
        scale = 1.0
    else:
        center = np.mean(points, axis=0)
        max_distance = max(np.linalg.norm(points - center, axis=1))
        scale = 10.0 * max_distance
    plt.figure()
    ax = plt.subplot(111, aspect="equal")
    for i, halfplane in enumerate(halfplanes):
        c = "r" if i < len(halfplanes) // 2 else "b"
        plot_halfplane(halfplane, ax, c, 0.5, scale)
    if points is not None:
        colors = ["r", "g", "b", "orange", "magenta", "brown", "k"][:len(points)]
        if len(colors) < len(points):
            colors.extend(["k"] * (len(points) - len(colors)))
        plt.scatter(points[:, 0], points[:, 1], c=colors, s=100)
    if xlim is not None:
        plt.xlim(xlim)
    if ylim is not None:
        plt.ylim(ylim)
    plt.show()


def plot_halfplane(halfplane, ax, c, alpha, scale):  # pragma: no cover
    line = (halfplane[:2] + np.linspace(-scale, scale, 101)[:, np.newaxis]
            * norm_vector(halfplane[2:]))
    ax.plot(line[:, 0], line[:, 1], lw=3, c=c, alpha=alpha)
    normal2d = np.array([-halfplane[3], halfplane[2]])
    for p in line[::10]:
        normal = (p + np.linspace(0.0, 0.1 * scale, 101)[:, np.newaxis]
                  * norm_vector(normal2d))
        ax.plot(normal[:, 0], normal[:, 1], c=c, alpha=0.5 * alpha)



"""GJK implementation of Jolt Physics.

Based on: A Fast and Robust GJK Implementation for Collision Detection of
Convex Objects - Gino van den Bergen,
http://www.dtecta.com/papers/jgt98convex.pdf

Copyright 2021 Jorrit Rouwe, MIT license
Source: https://github.com/jrouwe/JoltPhysics/blob/master/Jolt/Geometry/GJKClosestPoint.h
"""
import math
from enum import Enum

import numba
import numpy as np
from ..utils import EPSILON, MAX_FLOAT, scalar_triple_product


EPSILON_SQR = EPSILON * EPSILON
ALL_TRUE = np.array([True, True, True, True], dtype=np.dtype("bool"))


class GjkState(Enum):
    NoIntersection = 0
    Intersection = 1
    Unknown = 2
    Clipped = 3


def gjk_intersection_jolt(collider1, collider2, tolerance=1e-10):
    """Intersection test with Gilbert-Johnson-Keerthi (GJK) algorithm.

    This implementation differs in several ways from the libccd version:

    * Support points will be interpolated
    * A configurable numerical tolerance is used to check convergence
    * Simplex selection is done via an integer of which the first 4 bits are
      used to select or deselect points of the current simplex

    Implementation based on Jolt Physics, Copyright 2021 Jorrit Rouwe, MIT
    license.

    Based on: A Fast and Robust GJK Implementation for Collision Detection of
    Convex Objects - Gino van den Bergen,
    http://www.dtecta.com/papers/jgt98convex.pdf

    Parameters
    ----------
    collider1 : ConvexCollider
        Convex collider 1.

    collider2 : ConvexCollider
        Convex collider 2.

    tolerance : float, optional (default: 1e-10)
        Minimal distance between objects when the objects are considered to be
        colliding.

    Returns
    -------
    intersection : bool
        Do the two colliders intersect?
    """
    Y = np.empty((4, 3))  # Support points on A - B
    n_points = 0  # Number of points in Y that are valid

    tolerance_sq = tolerance * tolerance

    prev_v_len_sq = MAX_FLOAT
    search_direction = np.array([1.0, 0.0, 0.0])

    while True:
        # Get support points for shape A and B in search direction
        p = collider1.support_function(search_direction)
        q = collider2.support_function(-search_direction)
        state, n_points, prev_v_len_sq = _intersection_loop(
            p, q, Y, n_points, tolerance_sq, prev_v_len_sq, search_direction)
        if state == GjkState.Unknown:
            continue
        else:
            return state == GjkState.Intersection


@numba.njit(cache=True)
def _intersection_loop(
        p, q, Y, n_points, tolerance_sq, prev_v_len_sq, search_direction):
    # Get support point of the minkowski sum A - B of v
    support_point = p - q

    # If the support point is in the opposite direction as search_direction,
    # then we have found a separating axis and there is no intersection
    if search_direction.dot(support_point) < -EPSILON:
        # Separating axis found
        return GjkState.NoIntersection, n_points, prev_v_len_sq

    # Store the point for later use
    Y[n_points] = support_point
    n_points += 1

    # Determine the new closest point
    success, search_direction[:], v_len_sq, simplex = get_closest_point_to_origin(
        Y, n_points, prev_v_len_sq)
    if not success:
        return GjkState.NoIntersection, n_points, prev_v_len_sq

    # If there are 4 points, the origin is inside the tetrahedron and we're done
    if simplex == 0xf:
        return GjkState.Intersection, n_points, prev_v_len_sq

    # If v is very close to zero, we consider this a collision
    if v_len_sq <= tolerance_sq:
        return GjkState.Intersection, n_points, prev_v_len_sq

    # If v is very small compared to the length of y, we also consider this a
    # collision
    if v_len_sq <= EPSILON * max_y_length_squared(Y, n_points):
        return GjkState.Intersection, n_points, prev_v_len_sq

    # The next separation axis to test is the negative of the closest point of
    # the Minkowski sum to the origin.
    # Note: This must be done before terminating as converged since the
    # separating axis is -search_direction.
    search_direction *= -1.0

    # If the squared length of search_direction is not changing enough, we've
    # converged and there is no collision.
    assert prev_v_len_sq >= v_len_sq
    if prev_v_len_sq - v_len_sq <= EPSILON * prev_v_len_sq:
        # search_direction is a separating axis
        return GjkState.NoIntersection, n_points, prev_v_len_sq
    prev_v_len_sq = v_len_sq

    # Update the points of the simplex
    n_points = update_simplex_y(Y, n_points, simplex)

    return GjkState.Unknown, n_points, prev_v_len_sq


def gjk_distance_jolt(
        collider1, collider2, tolerance=1e-10, max_distance_squared=100000.0):
    """Gilbert-Johnson-Keerthi (GJK) algorithm for distance calculation.

    This implementation extends the intersection test by closest point
    calculation after convergence. It is also possible to clip colliders that
    are too far away early.

    Implementation based on Jolt Physics, Copyright 2021 Jorrit Rouwe, MIT
    license.

    Based on: A Fast and Robust GJK Implementation for Collision Detection of
    Convex Objects - Gino van den Bergen,
    http://www.dtecta.com/papers/jgt98convex.pdf

    Parameters
    ----------
    collider1 : ConvexCollider
        Convex collider 1.

    collider2 : ConvexCollider
        Convex collider 2.

    tolerance : float, optional (default: 1e-10)
        Minimal distance between objects when the objects are considered to be
        colliding.

    max_distance_squared : float, optional (default: 100000)
        The maximum squared distance between colliders before the objects are
        considered infinitely far away and processing is terminated.


    Returns
    -------
    distance : float
        The shortest distance between two convex shapes.

    closest_point1 : array, shape (3,)
        Closest point on first convex shape.
        If the distance is MAX_FLOAT the points are invalid.

    closest_point2 : array, shape (3,)
        Closest point on second convex shape.
        If the distance is MAX_FLOAT the points are invalid.

    simplex : array, shape (4, 3)
        Simplex defined by 4 points of the Minkowski difference between
        vertices of the two colliders.
    """
    Y = np.empty((4, 3))  # Support points on A - B
    P = np.empty((4, 3))  # Support points on A
    Q = np.empty((4, 3))  # Support points on B
    n_points = 0  # Number of points in Y, P and Q that are valid

    tolerance_sq = tolerance * tolerance

    search_direction = np.array([1.0, 0.0, 0.0])
    v_len_sq = np.dot(search_direction, search_direction)
    prev_v_len_sq = MAX_FLOAT

    while True:
        # Get support points for shape A and B in search direction
        p = collider1.support_function(search_direction)
        q = collider2.support_function(-search_direction)
        state, n_points, prev_v_len_sq, v_len_sq = _distance_loop(
            p, q, Y, P, Q, n_points, tolerance_sq, prev_v_len_sq, v_len_sq,
            search_direction, max_distance_squared)
        if state == GjkState.Unknown:
            continue
        elif state == GjkState.Clipped:
            return MAX_FLOAT, None, None, None
        else:
            # Get the closest points
            a, b = calculate_closest_points(Y, P, Q, n_points)

            assert abs(np.dot(search_direction, search_direction) - v_len_sq) < 1e-12
            dist = math.sqrt(v_len_sq)
            if dist < EPSILON:
                a = b = 0.5 * (a + b)
            return dist, a, b, Y


@numba.njit(cache=True)
def _distance_loop(
        p, q, Y, P, Q, n_points, tolerance_sq, prev_v_len_sq, v_len_sq,
        search_direction, max_distance_squared):
    # Get support point of the minkowski sum A - B of v
    support_point = p - q

    dot = search_direction.dot(support_point)

    # Test if we have a separation of more than max_distance_squared,
    # in which case we terminate early
    if dot < 0.0 and dot * dot > v_len_sq * max_distance_squared:
        return GjkState.Clipped, None, None, None

    # Store the point for later use
    Y[n_points] = support_point
    P[n_points] = p
    Q[n_points] = q
    n_points += 1

    success, ioV_new, v_len_sq_new, new_set = get_closest_point_to_origin(
        Y, n_points, prev_v_len_sq)
    if success:
        search_direction[:], v_len_sq, simplex = ioV_new, v_len_sq_new, new_set
    else:
        n_points -= 1  # Undo add last point
        simplex = 0b0000
        for i in range(n_points):
            simplex |= 1 << i

    # If there are 4 points, the origin is inside the tetrahedron,
    # and we're done
    if simplex == 0xf:
        v_len_sq = 0.0
        return GjkState.Intersection, n_points, prev_v_len_sq, v_len_sq

    # Update the points of the simplex
    n_points = update_simplex_ypq(Y, P, Q, n_points, simplex)

    # If v is very close to zero, we consider this a collision
    if v_len_sq <= tolerance_sq:
        v_len_sq = 0.0
        return GjkState.Intersection, n_points, prev_v_len_sq, v_len_sq

    # If v is very small compared to the length of y, we also consider
    # this a collision
    if v_len_sq <= EPSILON * max_y_length_squared(Y, n_points):
        v_len_sq = 0.0
        return GjkState.Intersection, n_points, prev_v_len_sq, v_len_sq

    # The next separation axis to test is the negative of the closest point of
    # the Minkowski sum to the origin
    # Note: This must be done before terminating as converged since the
    # separating axis is -v
    search_direction *= -1.0

    # If the squared length of v is not changing enough, we've converged and
    # there is no collision
    assert prev_v_len_sq >= v_len_sq
    if prev_v_len_sq - v_len_sq <= EPSILON * prev_v_len_sq:
        # search_direction is a separating axis
        return GjkState.NoIntersection, n_points, prev_v_len_sq, v_len_sq

    prev_v_len_sq = v_len_sq
    return GjkState.Unknown, n_points, prev_v_len_sq, v_len_sq


@numba.njit(cache=True)
def get_barycentric_coordinates_line(a, b):
    """Barycentric coordinates of the closest point to origin for infinite line.

    Point can then be computed as a * u + b * v.
    """
    ab = b - a
    denominator = np.dot(ab, ab)
    if denominator < EPSILON_SQR:
        # Degenerate line segment, fallback to points
        if np.dot(a, a) < np.dot(b, b):
            # A closest
            u = 1.0
            v = 0.0
        else:
            # B closest
            u = 0.0
            v = 1.0
    else:
        v = -a.dot(ab) / denominator
        u = 1.0 - v
    return u, v


@numba.njit(cache=True)
def get_barycentric_coordinates_plane(a, b, c):
    """Barycentric coordinates of closest point to origin for plane.

    Point can then be computed as a * u + b * v + c * w.
    Taken from: Real-Time Collision Detection - Christer Ericson
    (Section: Barycentric Coordinates) with p = 0.
    Adjusted to always include the shortest edge of the triangle in the
    calculation to improve numerical accuracy.
    """
    # First calculate the three edges
    v0 = b - a
    v1 = c - a
    v2 = c - b

    # Make sure that the shortest edge is included in the calculation to keep
    # the products a * b - c * d as small as possible to preserve accuracy
    d00 = v0.dot(v0)
    d11 = v1.dot(v1)
    d22 = v2.dot(v2)
    if d00 <= d22:
        # Use v0 and v1 to calculate barycentric coordinates
        d01 = v0.dot(v1)
        denominator = d00 * d11 - d01 * d01
        if abs(denominator) < EPSILON:
            # Degenerate triangle, return coordinates along longest edge
            if d00 > d11:
                u, v = get_barycentric_coordinates_line(a, b)
                w = 0.0
            else:
                u, w = get_barycentric_coordinates_line(a, c)
                v = 0.0
        else:
            a0 = a.dot(v0)
            a1 = a.dot(v1)
            v = (d01 * a1 - d11 * a0) / denominator
            w = (d01 * a0 - d00 * a1) / denominator
            u = 1.0 - v - w
    else:
        # Use v1 and v2 to calculate barycentric coordinates
        d12 = v1.dot(v2)

        denominator = d11 * d22 - d12 * d12
        if abs(denominator) < EPSILON:
            # Degenerate triangle, return coordinates along longest edge
            if d11 > d22:
                u, w = get_barycentric_coordinates_line(a, c)
                v = 0.0
            else:
                v, w = get_barycentric_coordinates_line(b, c)
                u = 0.0
        else:
            c1 = c.dot(v1)
            c2 = c.dot(v2)
            u = (d22 * c1 - d12 * c2) / denominator
            v = (d11 * c2 - d12 * c1) / denominator
            w = 1.0 - u - v
    return u, v, w


@numba.njit(cache=True)
def get_barycentric_coordinates_tetrahedron(a, b, c, d):
    """Barycentric coordinates of the closest point to origin for tetrahedron.

    Source: https://stackoverflow.com/a/38546111/915743
    """
    vab = b - a
    vac = c - a
    vad = d - a

    va6 = -scalar_triple_product(b, d - b, c - b)
    vb6 = -scalar_triple_product(a, vac, vad)
    vc6 = -scalar_triple_product(a, vad, vab)
    vd6 = -scalar_triple_product(a, vab, vac)
    v6 = 1.0 / scalar_triple_product(vab, vac, vad)
    return va6 * v6, vb6 * v6, vc6 * v6, vd6 * v6


@numba.njit(cache=True)
def closest_point_line(a, b):
    """Get the closest point to the origin of line.

    Output set describes which features are closest:

    * 1 = a
    * 2 = b
    * 3 = line segment ab
    """
    u, v = get_barycentric_coordinates_line(a, b)
    if v <= 0.0:
        # a is closest point
        return a, 0b0001
    elif u <= 0.0:
        # b is closest point
        return b, 0b0010
    else:
        # Closest point lies on line
        return u * a + v * b, 0b0011


@numba.njit(cache=True)
def closest_point_triangle(a, b, c):
    """Get the closest point to the origin of triangle.

    Output set describes which features are closest:

    * 1 = a
    * 2 = b
    * 4 = c
    * 5 = line segment ac
    * 7 = triangle interior etc.

    Taken from: Real-Time Collision Detection - Christer Ericson (Section:
    Closest Point on Triangle to Point) with p = 0.
    """
    # Calculate edges
    ab = b - a
    ac = c - a
    bc = c - b

    # The most accurate normal is calculated by using the two shortest edges
    # See: https://box2d.org/posts/2014/01/troublesome-triangle/
    # The difference in normals is most pronounced when one edge is much
    # smaller than the others (in which case the other 2 must have roughly the
    # same length). Therefore, we can suffice by just picking the shortest from
    # 2 edges and use that with the 3rd edge to calculate the normal. We first
    # check which of the edges is shorter.
    bc_shorter_than_ac = bc.dot(bc) < ac.dot(ac)
    if bc_shorter_than_ac:
        n = np.cross(ab, bc)
    else:
        n = np.cross(ab, ac)
    n_len_sq = np.dot(n, n)

    # Check degenerate
    if n_len_sq < EPSILON_SQR:
        # Degenerate, fallback to edges

        # Edge AB
        closest_point, closest_set = closest_point_line(a, b)
        best_dist_sq = np.dot(closest_point, closest_point)

        # Edge AC
        q, new_set = closest_point_line(a, c)
        dist_sq = np.dot(q, q)
        if dist_sq < best_dist_sq:
            closest_point = q
            best_dist_sq = dist_sq
            closest_set = (new_set & 0b0001) + ((new_set & 0b0010) << 1)

        # Edge BC
        q, new_set = closest_point_line(b, c)
        dist_sq = np.dot(q, q)
        if dist_sq < best_dist_sq:
            closest_point = q
            closest_set = new_set << 1

        return closest_point, closest_set

    # Check if P in vertex region outside A
    ap = -a
    d1 = ab.dot(ap)
    d2 = ac.dot(ap)
    if d1 <= 0.0 and d2 <= 0.0:
        return a, 0b0001  # barycentric coordinates (1,0,0)

    # Check if P in vertex region outside B
    bp = -b
    d3 = ab.dot(bp)
    d4 = ac.dot(bp)
    if d3 >= 0.0 and d4 <= d3:
        return b, 0b0010  # barycentric coordinates (0,1,0)

    # Check if P in edge region of AB, if so return projection of P onto AB
    vc = d1 * d4 - d3 * d2
    if vc <= 0.0 <= d1 and d3 <= 0.0:
        v = d1 / (d1 - d3)
        return a + v * ab, 0b0011  # barycentric coordinates (1-v,v,0)

    # Check if P in vertex region outside C
    cp = -c
    d5 = ab.dot(cp)
    d6 = ac.dot(cp)
    if d6 >= 0.0 and d5 <= d6:
        return c, 0b0100  # barycentric coordinates (0,0,1)

    # Check if P in edge region of AC, if so return projection of P onto AC
    vb = d5 * d2 - d1 * d6
    if vb <= 0.0 <= d2 and d6 <= 0.0:
        w = d2 / (d2 - d6)
        return a + w * ac, 0b0101  # barycentric coordinates (1-w,0,w)

    # Check if P in edge region of BC, if so return projection of P onto BC
    va = d3 * d6 - d5 * d4
    d4_d3 = d4 - d3
    d5_d6 = d5 - d6
    if va <= 0.0 <= d4_d3 and d5_d6 >= 0.0:
        w = d4_d3 / (d4_d3 + d5_d6)
        return b + w * bc, 0b0110  # barycentric coordinates (0,1-w,w)

    # P inside face region.
    # Here we deviate from Christer Ericson's article to improve accuracy.
    # Determine distance between triangle and origin:
    # distance = (centroid - origin) . normal / |normal|
    # Closest point to origin is then: distance . normal / |normal|
    # Note that this way of calculating the closest point is much more accurate
    # than first calculating barycentric coordinates and then calculating the
    # closest point based on those coordinates.
    return n * (a + b + c).dot(n) / (3.0 * n_len_sq), 0b0111


@numba.njit(cache=True)
def origin_outside_of_tetrahedron_planes(a, b, c, d):
    """Returns for each plane of the tetrahedron if the origin is inside.

    Roughly equivalent to:
       [OriginOutsideOfPlane(a, b, c, d),
        OriginOutsideOfPlane(a, c, d, b),
        OriginOutsideOfPlane(a, d, b, c),
        OriginOutsideOfPlane(b, d, c, a)]
    """
    ab = b - a
    ac = c - a
    ad = d - a
    bd = d - b
    bc = c - b

    ab_cross_ac = np.cross(ab, ac)
    ac_cross_ad = np.cross(ac, ad)
    ad_cross_ab = np.cross(ad, ab)
    bd_cross_bc = np.cross(bd, bc)

    # For each plane get the side on which the origin is
    signp0 = a.dot(ab_cross_ac)  # ABC
    signp1 = a.dot(ac_cross_ad)  # ACD
    signp2 = a.dot(ad_cross_ab)  # ADB
    signp3 = b.dot(bd_cross_bc)  # BDC
    signp = np.array([signp0, signp1, signp2, signp3])

    # For each plane get the side that is outside (determined by the 4th point)
    signd0 = ad.dot(ab_cross_ac)  # D
    signd1 = ab.dot(ac_cross_ad)  # B
    signd2 = ac.dot(ad_cross_ab)  # C
    signd3 = -ab.dot(bd_cross_bc)  # A
    signd = np.array([signd0, signd1, signd2, signd3])

    # The winding of all triangles has been chosen so that signd should have
    # the same sign for all components. If this is not the case the tetrahedron
    # is degenerate and we return that the origin is in front of all sides.
    if np.all(signd > 0.0):
        return signp >= -EPSILON
    elif np.all(signd < 0.0):
        return signp <= EPSILON
    else:
        # Mixed signs, degenerate tetrahedron
        return ALL_TRUE


@numba.njit(cache=True)
def closest_point_tetrahedron(a, b, c, d):
    """Get the closest point between tetrahedron to the origin.

    Output set specifies which feature was closest:

    * 1 = a
    * 2 = b
    * 4 = c
    * 8 = d

    Edges have 2 bits set, triangles 3, and if the point is in the interior
    4 bits are set.

    Taken from: Real-Time Collision Detection - Christer Ericson (Section:
    Closest Point on Tetrahedron to Point) with p = 0.
    """
    # Start out assuming point inside all halfspaces, so closest to itself
    closest_set = 0b1111
    closest_point = np.zeros(3)
    best_dist_sq = MAX_FLOAT

    # Determine for each of the faces of the tetrahedron if the origin is in
    # front of the plane
    origin_out_of_planes = origin_outside_of_tetrahedron_planes(a, b, c, d)

    # If point outside face abc then compute the closest point on abc
    if origin_out_of_planes[0]:  # OriginOutsideOfPlane(a, b, c, d)
        # Update best closest point because distance is less than MAX_FLOAT
        closest_point, closest_set = closest_point_triangle(a, b, c)
        best_dist_sq = np.dot(closest_point, closest_point)

    # Repeat test for face acd
    if origin_out_of_planes[1]:  # OriginOutsideOfPlane(a, c, d, b)
        q, new_set = closest_point_triangle(a, c, d)
        dist_sq = np.dot(q, q)
        if dist_sq < best_dist_sq:
            best_dist_sq = dist_sq
            closest_point = q
            closest_set = (new_set & 0b0001) + ((new_set & 0b0110) << 1)

    # Repeat test for face adb
    if origin_out_of_planes[2]:  # OriginOutsideOfPlane(a, d, b, c)
        q, new_set = closest_point_triangle(a, d, b)
        dist_sq = np.dot(q, q)
        if dist_sq < best_dist_sq:
            best_dist_sq = dist_sq
            closest_point = q
            closest_set = (new_set & 0b0001) + ((new_set & 0b0010) << 2) + ((new_set & 0b0100) >> 1)

    # Repeat test for face bdc
    if origin_out_of_planes[3]:  # OriginOutsideOfPlane(b, d, c, a)
        q, new_set = closest_point_triangle(b, d, c)
        dist_sq = np.dot(q, q)
        if dist_sq < best_dist_sq:
            closest_point = q
            closest_set = ((new_set & 0b0001) << 1) + ((new_set & 0b0010) << 2) + (new_set & 0b0100)

    return closest_point, closest_set


@numba.njit(cache=True)
def max_y_length_squared(y, n_points):
    """Get max(|Y_0|^2 .. |Y_n|^2)."""
    y_len_sq = np.dot(y[0], y[0])
    for i in range(1, n_points):
        y_len_sq = max(y_len_sq, np.dot(y[i], y[i]))
    return y_len_sq


@numba.njit(cache=True)
def update_simplex_y(Y, n_points, simplex):
    """Remove points that are not in the set, only updates Y."""
    n_new_points = 0
    for i in range(n_points):
        if (simplex & (1 << i)) != 0:
            Y[n_new_points] = Y[i]
            n_new_points += 1
    return n_new_points


@numba.njit(cache=True)
def update_simplex_ypq(Y, P, Q, n_points, simplex):
    """Remove points that are not in the set, updates Y, P and Q."""
    n_new_points = 0
    for i in range(n_points):
        if (simplex & (1 << i)) != 0:
            Y[n_new_points] = Y[i]
            P[n_new_points] = P[i]
            Q[n_new_points] = Q[i]
            n_new_points += 1
    return n_new_points


@numba.njit(cache=True)
def calculate_closest_points(Y, P, Q, n_points):
    """Calculate the closest points on A and B."""
    if n_points == 1:
        a = P[0]
        b = Q[0]
    elif n_points == 2:
        u, v = get_barycentric_coordinates_line(Y[0], Y[1])
        a = u * P[0] + v * P[1]
        b = u * Q[0] + v * Q[1]
    elif n_points == 3:
        u, v, w = get_barycentric_coordinates_plane(Y[0], Y[1], Y[2])
        a = u * P[0] + v * P[1] + w * P[2]
        b = u * Q[0] + v * Q[1] + w * Q[2]
    elif n_points == 4:  # intersection
        u, v, w, x = get_barycentric_coordinates_tetrahedron(Y[0], Y[1], Y[2], Y[3])
        a = u * P[0] + v * P[1] + w * P[2] + x * P[3]
        b = u * Q[0] + v * Q[1] + w * Q[2] + x * Q[3]
    else:
        a, b = None, None
    return a, b


@numba.njit(cache=True)
def get_closest_point_to_origin(Y, n_points, prev_v_len_sqr):
    """Get new closest point to origin given simplex Y of n_points points."""
    if n_points == 1:
        simplex = 0b0001
        v = Y[0]
    elif n_points == 2:
        v, simplex = closest_point_line(Y[0], Y[1])
    elif n_points == 3:
        v, simplex = closest_point_triangle(Y[0], Y[1], Y[2])
    elif n_points == 4:
        v, simplex = closest_point_tetrahedron(Y[0], Y[1], Y[2], Y[3])
    else:
        assert False

    v_len_sq = np.dot(v, v)
    # Note, comparison order important: If v_len_sq is NaN then this expression
    # will be False, so we will return False
    if v_len_sq < prev_v_len_sqr:
        return True, v, v_len_sq, simplex

    return False, None, None, None



import numpy as np
import numba
from ..utils import norm_vector, MAX_FLOAT, plane_basis_from_normal
from ..geometry import convert_segment_to_line, convert_rectangle_to_segment
from ._line import _line_to_line_segment
from ._rectangle import line_segment_to_rectangle


@numba.njit(numba.types.Tuple(
    (numba.float64, numba.float64[:]))(numba.float64[:], numba.float64[:, :]),
    cache=True)
def point_to_triangle(point, triangle_points):
    """Compute the shortest distance between point and triangle.

    Implementation adapted from Real-Time Collision Detection by Christer
    Ericson published by Morgan Kaufmann Publishers, Copyright 2005 Elsevier
    Inc.

    Parameters
    ----------
    point : array, shape (3,)
        3D point.

    triangle_points : array, shape (3, 3)
        Each row contains a point of the triangle (A, B, C).

    Returns
    -------
    distance : float
        The shortest distance between point and triangle.

    closest_point : array, shape (3,)
        Closest point on triangle.
    """
    ab = triangle_points[1] - triangle_points[0]
    ac = triangle_points[2] - triangle_points[0]

    # Check if point in vertex region outside A
    ap = point - triangle_points[0]
    d1 = np.dot(ab, ap)
    d2 = np.dot(ac, ap)
    if d1 <= 0.0 and d2 <= 0.0:
        closest_point = triangle_points[0]
        return np.linalg.norm(point - closest_point), closest_point

    # Check if point in vertex region outside B
    bp = point - triangle_points[1]
    d3 = np.dot(ab, bp)
    d4 = np.dot(ac, bp)
    if d3 >= 0 and d4 <= d3:
        closest_point = triangle_points[1]
        return np.linalg.norm(point - closest_point), closest_point

    # Check if point in edge region of AB
    vc = d1 * d4 - d3 * d2
    if vc <= 0.0 <= d1 and d3 <= 0.0:
        v = d1 / (d1 - d3)
        closest_point = triangle_points[0] + v * ab
        return np.linalg.norm(point - closest_point), closest_point

    # Check if point in vertex region outside C
    cp = point - triangle_points[2]
    d5 = np.dot(ab, cp)
    d6 = np.dot(ac, cp)
    if d6 >= 0.0 and d5 <= d6:
        closest_point = triangle_points[2]
        return np.linalg.norm(point - closest_point), closest_point

    # Check if point in edge region of AC
    vb = d5 * d2 - d1 * d6
    if vb <= 0.0 <= d2 and d6 <= 0.0:
        w = d2 / (d2 - d6)
        closest_point = triangle_points[0] + w * ac
        return np.linalg.norm(point - closest_point), closest_point

    # Check if point in edge region of BC
    va = d3 * d6 - d5 * d4
    if va <= 0.0 <= d4 - d3 and d5 - d6 >= 0.0:
        w = (d4 - d3) / ((d4 - d3) + (d5 - d6))
        closest_point = triangle_points[1] + w * (triangle_points[2] - triangle_points[1])
        return np.linalg.norm(point - closest_point), closest_point

    # Point inside face region
    denom = 1.0 / (va + vb + vc)
    v = vb * denom
    w = vc * denom
    closest_point = triangle_points[0] + ab * v + ac * w

    return np.linalg.norm(point - closest_point), closest_point


def line_to_triangle(line_point, line_direction, triangle_points, epsilon=1e-6):
    """Compute the shortest distance between point and triangle.

    Implementation adapted from Real-Time Collision Detection by Christer
    Ericson published by Morgan Kaufmann Publishers, Copyright 2005 Elsevier
    Inc.

    Parameters
    ----------
    line_point : array, shape (3,)
        Point on line.

    line_direction : array, shape (3,)
        Direction of the line. This is assumed to be of unit length.

    triangle_points : array, shape (3, 3)
        Each row contains a point of the triangle (A, B, C).

    epsilon : float, optional (default: 1e-6)
        Values smaller than epsilon are considered to be 0.

    Returns
    -------
    distance : float
        The shortest distance between line and triangle.

    closest_point_line : array, shape (3,)
        Closest point on line.

    closest_point_triangle : array, shape (3,)
        Closest point on triangle.
    """
    return _line_to_triangle(line_point, line_direction, triangle_points, epsilon)[:3]


@numba.njit(numba.types.Tuple(
    (numba.float64, numba.float64[:], numba.float64[:], numba.float64))
    (numba.float64[::1], numba.float64[::1], numba.float64[:, ::1], numba.float64),
    cache=True)
def _line_to_triangle(line_point, line_direction, triangle_points, epsilon):
    # Test if line intersects triangle. If so, the squared distance is zero.
    edge = np.row_stack((triangle_points[1] - triangle_points[0],
                         triangle_points[2] - triangle_points[0]))
    normal = norm_vector(np.cross(edge[0], edge[1]))
    if abs(normal.dot(line_direction)) > epsilon:
        # The line and triangle are not parallel, so the line intersects
        # the plane of the triangle.
        diff = line_point - triangle_points[0]
        u, v = plane_basis_from_normal(line_direction)
        ude = edge.dot(u)
        vde = edge.dot(v)
        uddiff = u.dot(diff)
        vddiff = v.dot(diff)
        det = ude[0] * vde[1] - ude[1] * vde[0]

        # Barycentric coordinates for the point of intersection.
        b = np.array([vde[1] * uddiff - ude[1] * vddiff,
                      ude[0] * vddiff - vde[0] * uddiff])
        if det != 0.0:
            b /= det
        b2 = 1.0 - b[0] - b[1]

        if b2 >= 0.0 and b[0] >= 0.0 and b[1] >= 0.0:
            # Line parameter for the point of intersection.
            dde = edge.dot(line_direction)
            dddiff = line_direction.dot(diff)
            line_parameter = b.dot(dde) - dddiff

            # The intersection point is inside or on the triangle.
            best_closest_point_line = line_point + line_parameter * line_direction
            best_closest_point_triangle = triangle_points[0] + b.dot(edge)
            return 0.0, best_closest_point_line, best_closest_point_triangle, line_parameter

    # Either (1) the line is not parallel to the triangle and the point of
    # intersection of the line and the plane of the triangle is outside the
    # triangle or (2) the line and triangle are parallel. Regardless, the
    # closest point on the triangle is on an edge of the triangle. Compare
    # the line to all three edges of the triangle.
    best_distance = MAX_FLOAT
    i0 = 2
    i1 = 0
    while i1 < 3:
        distance, closest_point_line, closest_point_segment, t, _ = _line_to_line_segment(
            line_point, line_direction, triangle_points[i0], triangle_points[i1],
            epsilon)

        if distance < best_distance:
            best_closest_point_line = closest_point_line
            best_closest_point_triangle = closest_point_segment
            best_distance = distance
            best_line_parameter = t

        i0 = i1
        i1 += 1
    return best_distance, best_closest_point_line, best_closest_point_triangle, best_line_parameter


def line_segment_to_triangle(
        segment_start, segment_end, triangle_points, epsilon=1e-6):
    """Compute the shortest distance from line segment to triangle.

    Implementation adapted from Real-Time Collision Detection by Christer
    Ericson published by Morgan Kaufmann Publishers, Copyright 2005 Elsevier
    Inc.

    Parameters
    ----------
    segment_start : array, shape (3,)
        Start point of segment.

    segment_end : array, shape (3,)
        End point of segment.

    triangle_points : array, shape (3, 3)
        Each row contains a point of the triangle (A, B, C).

    epsilon : float, optional (default: 1e-6)
        Values smaller than epsilon are considered to be 0.

    Returns
    -------
    distance : float
        The shortest distance between line segment and triangle.

    closest_point_line_segment : array, shape (3,)
        Closest point on line segment.

    closest_point_triangle : array, shape (3,)
        Closest point on triangle.
    """
    segment_direction, segment_length = convert_segment_to_line(
        segment_start, segment_end)

    distance, closest_point_segment, closest_point_triangle, t_closest = _line_to_triangle(
        segment_start, segment_direction, triangle_points, epsilon)

    if t_closest < 0:
        distance, closest_point_triangle = point_to_triangle(
            segment_start, triangle_points)
        closest_point_segment = segment_start
    elif t_closest > segment_length:
        distance, closest_point_triangle = point_to_triangle(
            segment_end, triangle_points)
        closest_point_segment = segment_end

    return distance, closest_point_segment, closest_point_triangle


def triangle_to_triangle(triangle_points1, triangle_points2, epsilon=1e-6):
    """Compute the shortest distance between two triangles.

    Implementation adapted from 3D Game Engine Design by David H. Eberly.

    Geometric Tools, Inc.
    http://www.geometrictools.com
    Copyright (c) 1998-2006.  All Rights Reserved

    The Wild Magic Version 4 Foundation Library source code is supplied
    under the terms of the license agreement
    (http://www.geometrictools.com/License/Wm4FoundationLicense.pdf)
    and may not be copied or disclosed except in accordance with the terms
    of that agreement.

    Parameters
    ----------
    triangle_points1 : array, shape (3, 3)
        Each row contains a point of the first triangle.

    triangle_points2 : array, shape (3, 3)
        Each row contains a point of the second triangle.

    epsilon : float, optional (default: 1e-6)
        Values smaller than epsilon are considered to be 0.

    Returns
    -------
    distance : float
        Shortest distance.

    closest_point_triangle1 : array, shape (3,)
        Closest point on triangle 1.

    closest_point_triangle2 : array, shape (3,)
        Closest point on triangle 2.
    """
    # compare edges of triangle1 to the interior of triangle2
    best_dist = MAX_FLOAT
    i0 = 2
    i1 = 0
    while i1 < 3:
        segment_start = triangle_points1[i0]
        segment_end = triangle_points1[i1]
        dist, closest_point_segment, closest_point_triangle = line_segment_to_triangle(
            segment_start, segment_end, triangle_points2, epsilon)
        if dist < best_dist:
            best_closest_point_triangle1 = closest_point_segment
            best_closest_point_triangle2 = closest_point_triangle
            best_dist = dist

            if best_dist <= epsilon:
                return 0.0, best_closest_point_triangle1, best_closest_point_triangle2
        i0 = i1
        i1 += 1

    # compare edges of triangle2 to the interior of triangle1
    i0 = 2
    i1 = 0
    while i1 < 3:
        segment_start = triangle_points2[i0]
        segment_end = triangle_points2[i1]
        dist, closest_point_segment, closest_point_triangle = line_segment_to_triangle(
            segment_start, segment_end, triangle_points1, epsilon)
        if dist < best_dist:
            best_closest_point_triangle1 = closest_point_triangle
            best_closest_point_triangle2 = closest_point_segment
            best_dist = dist

            if best_dist <= epsilon:
                return 0.0, best_closest_point_triangle1, best_closest_point_triangle2
        i0 = i1
        i1 += 1

    return best_dist, best_closest_point_triangle1, best_closest_point_triangle2


def triangle_to_rectangle(
        triangle_points, rectangle_center, rectangle_axes, rectangle_lengths):
    """Compute the shortest distance between triangle and rectangle.

    Parameters
    ----------
    triangle_points : array, shape (3, 3)
        Each row contains a point of the triangle.

    rectangle_center : array, shape (3,)
        Center point of the rectangle.

    rectangle_axes : array, shape (2, 3)
        Each row is a vector of unit length, indicating the direction of one
        axis of the rectangle. Both vectors are orthogonal.

    rectangle_lengths : array, shape (2,)
        Lengths of the two sides of the rectangle.

    Returns
    -------
    distance : float
        Shortest distance.

    closest_point_triangle : array, shape (3,)
        Closest point on triangle.

    closest_point_rectangle : array, shape (3,)
        Closest point on rectangle.
    """
    # compare edges of triangle to the interior of rectangle
    best_dist = MAX_FLOAT
    i0 = 2
    i1 = 0
    while i1 < 3:
        segment_start = triangle_points[i0]
        segment_end = triangle_points[i1]
        dist, closest_point_triangle, closest_point_rectangle = line_segment_to_rectangle(
            segment_start, segment_end, rectangle_center, rectangle_axes, rectangle_lengths)
        if dist < best_dist:
            best_closest_point_triangle = closest_point_triangle
            best_closest_point_rectangle = closest_point_rectangle
            best_dist = dist
        i0 = i1
        i1 += 1

    # compare edges of rectangle to the interior of triangle
    rectangle_extents = 0.5 * rectangle_lengths[:, np.newaxis] * rectangle_axes
    for i1 in range(2):
        for i0 in range(2):
            segment_end, segment_start = convert_rectangle_to_segment(
                rectangle_center, rectangle_extents, i0, i1)
            dist, closest_point_rectangle, closest_point_triangle = line_segment_to_triangle(
                segment_start, segment_end, triangle_points)
            if dist < best_dist:
                best_closest_point_triangle = closest_point_triangle
                best_closest_point_rectangle = closest_point_rectangle
                best_dist = dist

    return best_dist, best_closest_point_triangle, best_closest_point_rectangle



import math
from enum import Enum
import numba
import numpy as np

from distance3d.utils import EPSILON
from distance3d.minkowski import Simplex, support_function, make_support_point
from distance3d.distance import point_to_triangle


EPSILON_SQRT = math.sqrt(EPSILON)


def gjk_intersection_libccd(collider1, collider2, max_iterations=100):
    """Intersection test with Gilbert-Johnson-Keerthi (GJK) algorithm.

    This version of GJK does not interpolate between vertices of a mesh.
    Hence, it might sometimes loop through the same vertices of a mesh
    until the maximum number of iterations is reached. This does not occur
    with parametrically represented primitive shapes.

    An advantage of the implementation is that it implements the intersection
    test much faster than the original implementation. It is also the shortest
    implementation of all.

    This implementation of GJK is based on libccd (for details, see
    https://github.com/danfis/libccd). For the original code the copyright is
    of Daniel Fiser <danfis@danfis.cz>. It has been released under 3-clause BSD
    license.

    Based on: A Fast and Robust GJK Implementation for Collision Detection of
    Convex Objects - Gino van den Bergen,
    http://www.dtecta.com/papers/jgt98convex.pdf

    Parameters
    ----------
    collider1 : ConvexCollider
        Convex collider 1.

    collider2 : ConvexCollider
        Convex collider 2.

    max_iterations : int, optional (default: 100)
        Maximum number of iterations.

    Returns
    -------
    intersection : bool
        Do the two colliders intersect?
    """
    return _gjk(collider1, collider2, Simplex(), max_iterations)[0]


def _gjk(collider1, collider2, simplex, max_iterations):
    support_point = make_support_point(
        collider1.first_vertex(), collider2.first_vertex())
    simplex.add_point(*support_point)
    search_direction = -support_point[0]

    for _ in range(max_iterations):
        support_point = support_function(collider1, collider2, search_direction)
        support_point_is_origin = np.dot(
            support_point[0], support_point[0]) < EPSILON
        if support_point_is_origin:
            _set_point(simplex.v, simplex.v1, simplex.v2, 0, *support_point)
            simplex.n_points = 1
            return True, simplex

        support_point_is_before_origin = np.dot(
            support_point[0], search_direction) < -EPSILON_SQRT
        if support_point_is_before_origin:
            return False, simplex

        simplex.add_point(*support_point)

        gjk_state, search_direction, simplex.n_points = _refine_simplex(
            simplex.v, simplex.v1, simplex.v2, len(simplex))

        if gjk_state == GjkState.CONTACT:
            return True, simplex
        elif gjk_state == GjkState.NO_CONTACT:
            return False, simplex
        elif abs(np.dot(search_direction, search_direction)) < EPSILON:
            return False, simplex

    return False, simplex


class GjkState(Enum):
    NO_CONTACT = -1
    CONTINUE = 0
    CONTACT = 1


@numba.njit(cache=True)
def _refine_simplex(v, v1, v2, n_points):
    if n_points == 2:
        return _line_segment(v, v1, v2)
    elif n_points == 3:
        return _triangle(v, v1, v2)
    else:
        return _tetrahedron(v, v1, v2)


@numba.njit(cache=True)
def _set_point(v, v1, v2, idx, new_v, new_v1, new_v2):
    v[idx] = new_v
    v1[idx] = new_v1
    v2[idx] = new_v2


@numba.njit(cache=True)
def _line_segment(v, v1, v2):
    A = v[1], v1[1], v2[1]
    B = v[0]

    AB = B - A[0]
    AO = -A[0]
    origin_on_AB = np.dot(AB, AO)

    tmp = np.cross(AB, AO)
    origin_on_AB_segment = (
        abs(np.dot(tmp, tmp)) < EPSILON and origin_on_AB > 0.0)
    if origin_on_AB_segment:
        return GjkState.CONTACT, None, 2

    if origin_on_AB < EPSILON:
        _set_point(v, v1, v2, 0, *A)
        n_points = 1
        search_direction = AO
    else:  # origin is closer to line segment
        search_direction = _triple_cross(AB, AO, AB)
        n_points = 2

    return GjkState.CONTINUE, search_direction, n_points


@numba.njit(cache=True)
def _triangle(v, v1, v2):
    A = np.copy(v[2]), np.copy(v1[2]), np.copy(v1[2])
    B = np.copy(v[1]), np.copy(v1[1]), np.copy(v2[1])
    C = np.copy(v[0]), np.copy(v1[0]), np.copy(v1[0])

    touching_contact = abs(point_to_triangle(
        np.zeros(3), np.row_stack((A[0], B[0], C[0])))[0]) < EPSILON_SQRT
    if touching_contact:
        return GjkState.CONTACT, None, 1

    degenerated_triangle = (np.all(np.abs(A[0] - B[0]) < EPSILON)
                            or np.all(np.abs(A[0] - C[0]) < EPSILON))
    if degenerated_triangle:
        return GjkState.NO_CONTACT, None, 0

    AO = -A[0]

    AB = B[0] - A[0]
    AC = C[0] - A[0]
    ABC = np.cross(AB, AC)

    if np.dot(np.cross(ABC, AC), AO) > -EPSILON:
        if np.dot(AC, AO) > -EPSILON:
            _set_point(v, v1, v2, 1, *A)
            n_points = 2
            search_direction = _triple_cross(AC, AO, AC)
        else:
            n_points, search_direction = _triangle_ab(A, B, AB, AO, v, v1, v2)
    else:
        if np.dot(np.cross(AB, ABC), AO) > -EPSILON:
            n_points, search_direction = _triangle_ab(A, B, AB, AO, v, v1, v2)
        else:
            if np.dot(ABC, AO) > -EPSILON:
                n_points = 3
                search_direction = ABC
            else:
                _set_point(v, v1, v2, 0, *B)
                _set_point(v, v1, v2, 1, *C)
                n_points = 3
                search_direction = -ABC

    return GjkState.CONTINUE, search_direction, n_points


@numba.njit(cache=True)
def _triangle_ab(A, B, AB, AO, v, v1, v2):
    if np.dot(AB, AO) > -EPSILON:
        _set_point(v, v1, v2, 0, *B)
        _set_point(v, v1, v2, 1, *A)
        n_points = 2
        search_direction = _triple_cross(AB, AO, AB)
    else:
        _set_point(v, v1, v2, 0, *A)
        n_points = 1
        search_direction = AO
    return n_points, search_direction


@numba.njit(cache=True)
def _tetrahedron(v, v1, v2):
    A = np.copy(v[3]), np.copy(v1[3]), np.copy(v1[3])
    B = np.copy(v[2]), np.copy(v1[2]), np.copy(v1[2])
    C = np.copy(v[1]), np.copy(v1[1]), np.copy(v2[1])
    D = np.copy(v[0]), np.copy(v1[0]), np.copy(v1[0])

    degenerated_tetrahedron = abs(point_to_triangle(
        A[0], np.row_stack((B[0], C[0], D[0])))[0]) < EPSILON_SQRT
    if degenerated_tetrahedron:
        return GjkState.NO_CONTACT, None, 0

    origin = np.zeros(3)
    origin_lies_on_tetrahedrons_face = (
        point_to_triangle(origin, np.row_stack((A[0], B[0], C[0])))[0] < EPSILON_SQRT
        or point_to_triangle(origin, np.row_stack((A[0], C[0], D[0])))[0] < EPSILON_SQRT
        or point_to_triangle(origin, np.row_stack((A[0], B[0], D[0])))[0] < EPSILON_SQRT
        or point_to_triangle(origin, np.row_stack((B[0], C[0], D[0])))[0] < EPSILON_SQRT
    )
    if origin_lies_on_tetrahedrons_face:
        return GjkState.CONTACT, None, 3

    # compute AO, AB, AC, AD segments and ABC, ACD, ADB normal vectors
    AO = -A[0]
    AB = B[0] - A[0]
    AC = C[0] - A[0]
    AD = D[0] - A[0]
    ABC = np.cross(AB, AC)
    ACD = np.cross(AC, AD)
    ADB = np.cross(AD, AB)

    # side (positive or negative) of B, C, D relative to planes ACD, ADB
    # and ABC respectively
    B_on_ACD = np.sign(np.dot(ACD, AB))
    C_on_ADB = np.sign(np.dot(ADB, AC))
    D_on_ABC = np.sign(np.dot(ABC, AD))

    # whether origin is on same side of ACD, ADB, ABC as B, C, D
    # respectively
    AB_O = np.sign(np.dot(ACD, AO)) == B_on_ACD
    AC_O = np.sign(np.dot(ADB, AO)) == C_on_ADB
    AD_O = np.sign(np.dot(ABC, AO)) == D_on_ABC

    origin_is_in_tetrahedron = AB_O and AC_O and AD_O
    if origin_is_in_tetrahedron:
        return GjkState.CONTACT, None, 4

    _rearrange_simplex_to_triangle(A, B, C, D, AB_O, AC_O, v, v1, v2)

    return _triangle(v, v1, v2)


@numba.njit(cache=True)
def _rearrange_simplex_to_triangle(A, B, C, D, AB_O, AC_O, v, v1, v2):
    if not AB_O:
        _set_point(v, v1, v2, 2, *A)
    elif not AC_O:
        _set_point(v, v1, v2, 1, *D)
        _set_point(v, v1, v2, 0, *B)
        _set_point(v, v1, v2, 2, *A)
    else:
        _set_point(v, v1, v2, 0, *C)
        _set_point(v, v1, v2, 1, *B)
        _set_point(v, v1, v2, 2, *A)


@numba.njit(cache=True)
def _triple_cross(a, b, c):
    """d = a x b x c"""
    return np.cross(np.cross(a, b), c)



import enum
import math
import numpy as np


def make_triangular_icosphere(center, radius, order=4):
    """Create a triangular icosphere mesh.

    Source: https://observablehq.com/@mourner/fast-icosphere-mesh

    Parameters
    ----------
    center : array, shape (3,)
        Center of the sphere.

    radius : float
        Radius of the sphere.

    order : int, optional (default: 4)
        Number of subdivisions of initial 20 triangles.

    Returns
    -------
    vertices : array, shape (n_vertices, 3)
        Vertices of the mesh.

    triangles : array, shape (n_triangles, 3)
        Indices of vertices that form triangles of the mesh.
    """
    f = (1 + 5 ** 0.5) / 2
    vertices = np.zeros(((10 * 4 ** order + 2), 3))
    vertices[:12] = np.array([
        [-1, f, 0], [1, f, 0], [-1, -f, 0],
        [1, -f, 0], [0, -1, f], [0, 1, f],
        [0, -1, -f], [0, 1, -f], [f, 0, -1],
        [f, 0, 1], [-f, 0, -1], [-f, 0, 1]])
    triangles = np.array([
        [0, 11, 5], [0, 5, 1], [0, 1, 7], [0, 7, 10], [0, 10, 11], [11, 10, 2],
        [5, 11, 4], [1, 5, 9], [7, 1, 8], [10, 7, 6], [3, 9, 4], [3, 4, 2],
        [3, 2, 6], [3, 6, 8], [3, 8, 9], [9, 8, 1], [4, 9, 5], [2, 4, 11],
        [6, 2, 10], [8, 6, 7]
    ], dtype=int)
    v = 12
    # midpoint vertices cache to avoid duplicating shared vertices
    mid_cache = dict()

    def add_mid_point(a, b, mid_cache, v):
        # Cantor's pairing function
        key = math.floor((a + b) * (a + b + 1) / 2) + min(a, b)
        i = mid_cache.get(key, None)
        if i is not None:
            del mid_cache[key]
            return i, v
        mid_cache[key] = v
        vertices[v] = 0.5 * (vertices[a] + vertices[b])
        i = v
        v += 1
        return i, v

    # repeatedly subdivide each triangle into 4 triangles
    triangles_prev = triangles
    for _ in range(order):
        triangles = np.empty(
            (4 * triangles.shape[0], triangles.shape[1]),
            dtype=int)
        for k, triangle in enumerate(triangles_prev):
            v1, v2, v3 = triangle
            a, v = add_mid_point(v1, v2, mid_cache, v)
            b, v = add_mid_point(v2, v3, mid_cache, v)
            c, v = add_mid_point(v3, v1, mid_cache, v)
            t = k * 4
            triangles[t] = v1, a, c
            triangles[t + 1] = v2, b, a
            triangles[t + 2] = v3, c, b
            triangles[t + 3] = a, b, c
        triangles_prev = triangles

    vertices /= 1.0 / radius * np.linalg.norm(vertices, axis=1)[:, np.newaxis]
    vertices += center[np.newaxis]
    return vertices, triangles


def make_tetrahedral_sphere(radius, order=4):
    """Create a tetrahedral icosphere mesh.

    Parameters
    ----------
    radius : float
        Radius of the sphere.

    order : int, optional (default: 4)
        Number of subdivisions of initial 20 triangles.

    Returns
    -------
    vertices : array, shape (n_vertices, 3)
        Vertices of the mesh.

    tetrahedra : array, shape (n_tetrahedra, 4)
        Indices of vertices that form tetrahedra of the mesh.

    potentials : array, shape (n_vertices, 3)
        Potential of each vertex.
    """
    vertices, triangles = make_triangular_icosphere(np.zeros(3), radius, order)
    center_idx = len(vertices)
    vertices = np.vstack((vertices, np.zeros((1, 3))))
    tetrahedra = np.hstack(
        (triangles, center_idx * np.ones((len(triangles), 1), dtype=int)))
    potentials = np.zeros(len(vertices))
    potentials[-1] = radius
    return vertices, tetrahedra, potentials


def make_tetrahedral_ellipsoid(radii, order=4):
    """Create a tetrahedral ellipsoid mesh.

    Parameters
    ----------
    radii : array, shape (3,)
        Radii of the ellipsoid.

    order : int, optional (default: 4)
        Number of subdivisions of initial 20 triangles.

    Returns
    -------
    vertices : array, shape (n_vertices, 3)
        Vertices of the mesh.

    tetrahedra : array, shape (n_tetrahedra, 4)
        Indices of vertices that form tetrahedra of the mesh.

    potentials : array, shape (n_vertices, 3)
        Potential of each vertex.
    """
    vertices, triangles = make_triangular_icosphere(np.zeros(3), 1.0, order)
    vertices *= radii[np.newaxis]
    center_idx = len(vertices)
    vertices = np.vstack((vertices, np.zeros((1, 3))))
    tetrahedra = np.hstack(
        (triangles, center_idx * np.ones((len(triangles), 1), dtype=int)))
    potentials = np.zeros(len(vertices))
    potentials[-1] = min(radii)
    return vertices, tetrahedra, potentials


def make_tetrahedral_cube(size):
    """Create a tetrahedral cube mesh.

    Parameters
    ----------
    size : float
        Length of the edges in each dimension.

    Returns
    -------
    vertices : array, shape (n_vertices, 3)
        Vertices of the mesh.

    tetrahedra : array, shape (n_tetrahedra, 4)
        Indices of vertices that form tetrahedra of the mesh.

    potentials : array, shape (n_vertices, 3)
        Potential of each vertex.
    """
    vertices = size * np.array([
        [-0.5, -0.5, -0.5],
        [-0.5, -0.5, 0.5],
        [-0.5, 0.5, -0.5],
        [-0.5, 0.5, 0.5],
        [0.5, -0.5, -0.5],
        [0.5, -0.5, 0.5],
        [0.5, 0.5, -0.5],
        [0.5, 0.5, 0.5],
        [0.0, 0.0, 0.0]
    ])
    tetrahedra = np.array([
        [0, 2, 6, 8],
        [0, 4, 5, 8],
        [0, 1, 2, 8],
        [1, 3, 2, 8],
        [1, 5, 7, 8],
        [1, 7, 3, 8],
        [5, 1, 0, 8],
        [5, 6, 7, 8],
        [6, 2, 3, 8],
        [6, 3, 7, 8],
        [6, 4, 0, 8],
        [6, 5, 4, 8],
    ], dtype=int)
    potentials = np.array(
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, size / 2.0],
        dtype=float)
    return vertices, tetrahedra, potentials


def make_tetrahedral_box(size):
    """Create a tetrahedral box mesh.

    Source: Drake (https://github.com/RobotLocomotion/drake/blob/6b4664c2b4c4a7f52d24d16898d0bc2cc7f2b893/geometry/proximity/make_box_mesh.cc#L245),
    BSD 3-clause

    Parameters
    ----------
    size : array, shape (3,)
        Lengths of the edges in each dimension.

    Returns
    -------
    vertices : array, shape (n_vertices, 3)
        Vertices of the mesh.

    tetrahedra : array, shape (n_tetrahedra, 4)
        Indices of vertices that form tetrahedra of the mesh.

    potentials : array, shape (n_vertices, 3)
        Potential of each vertex.
    """
    mesh_vertices = []
    v = np.empty((2, 2, 2), dtype=float)
    half_size = 0.5 * size
    for i in range(2):
        x = -half_size[0] if i == 0 else half_size[0]
        for j in range(2):
            y = -half_size[1] if j == 0 else half_size[1]
            for k in range(2):
                z = -half_size[2] if k == 0 else half_size[2]
                v[i, j, k] = len(mesh_vertices)
                mesh_vertices.append([x, y, z])
    m = np.empty((2, 2, 2), dtype=float)
    min_half_size = min(half_size)
    relative_tolerance = 1e-14 * max(1.0, min_half_size)
    half_central = half_size - min_half_size
    half_central[half_central <= relative_tolerance] = 0.0
    for i in range(2):
        x = -half_central[0] if i == 0 else half_central[0]
        for j in range(2):
            y = -half_central[1] if j == 0 else half_central[1]
            for k in range(2):
                z = -half_central[2] if k == 0 else half_central[2]
                duplicate_in_i = i == 1 and half_central[0] == 0.0
                duplicate_in_j = j == 1 and half_central[1] == 0.0
                duplicate_in_k = k == 1 and half_central[2] == 0.0
                if duplicate_in_i:
                    m[i, j, k] = m[0, j, k]
                elif duplicate_in_j:
                    m[i, j, k] = m[i, 0, k]
                elif duplicate_in_k:
                    m[i, j, k] = m[i, j, 0]
                else:
                    m[i, j, k] = len(mesh_vertices)

                if not duplicate_in_i and not duplicate_in_j and not duplicate_in_k:
                    mesh_vertices.append([x, y, z])
    mesh_vertices = np.array(mesh_vertices)
    assert len(mesh_vertices) <= 12

    mesh_elements = []
    mesh_elements.extend(_split_to_tetrahedra(
        m[1, 0, 0], m[1, 1, 0], m[1, 1, 1], m[1, 0, 1],
        v[1, 0, 0], v[1, 1, 0], v[1, 1, 1], v[1, 0, 1]
    ))
    mesh_elements.extend(_split_to_tetrahedra(
        m[0, 0, 0], m[0, 0, 1], m[0, 1, 1], m[0, 1, 0],
        v[0, 0, 0], v[0, 0, 1], v[0, 1, 1], v[0, 1, 0]
    ))
    mesh_elements.extend(_split_to_tetrahedra(
        m[0, 1, 0], m[0, 1, 1], m[1, 1, 1], m[1, 1, 0],
        v[0, 1, 0], v[0, 1, 1], v[1, 1, 1], v[1, 1, 0]
    ))
    mesh_elements.extend(_split_to_tetrahedra(
        m[0, 0, 0], m[1, 0, 0], m[1, 0, 1], m[0, 0, 1],
        v[0, 0, 0], v[1, 0, 0], v[1, 0, 1], v[0, 0, 1]
    ))
    mesh_elements.extend(_split_to_tetrahedra(
        m[0, 0, 1], m[1, 0, 1], m[1, 1, 1], m[0, 1, 1],
        v[0, 0, 1], v[1, 0, 1], v[1, 1, 1], v[0, 1, 1]
    ))
    mesh_elements.extend(_split_to_tetrahedra(
        m[0, 0, 0], m[0, 1, 0], m[1, 1, 0], m[1, 0, 0],
        v[0, 0, 0], v[0, 1, 0], v[1, 1, 0], v[1, 0, 0]
    ))
    mesh_elements = np.array(mesh_elements, dtype=int)

    potentials = np.zeros(len(mesh_vertices))
    potentials[8:] = min_half_size

    return mesh_vertices, mesh_elements, potentials


def _split_to_tetrahedra(v0, v1, v2, v3, v4, v5, v6, v7):
    elements = []
    previous = v1
    for next in [v2, v3, v7, v4, v5, v1]:
        if len({previous, next, v0, v6}) == 4:
            elements.append([previous, next, v0, v6])
        previous = next
    return elements


def make_tetrahedral_cylinder(radius, length, resolution_hint):
    """Create a tetrahedral cylinder mesh.

    Source: Drake (https://github.com/RobotLocomotion/drake/blob/665f178d7c61edef1a4961cc44bb320062224944/geometry/proximity/make_cylinder_mesh.cc#L345),
    BSD 3-clause

    Parameters
    ----------
    radius : float
        Radius of the cylinder.

    length : float
        Length of the cylinder.

    resolution_hint : float
        Controls the fineness of the tetrahedral mesh. The coarsest mesh
        that produces desirable results will allow simulation to run as
        efficiently as possible. The circles of the cylinder will have
        2 * pi * radius / resolution_hint edges.

    Returns
    -------
    vertices : array, shape (n_vertices, 3)
        Vertices of the mesh.

    tetrahedra : array, shape (n_tetrahedra, 4)
        Indices of vertices that form tetrahedra of the mesh.

    potentials : array, shape (n_vertices, 3)
        Potential of each vertex.
    """
    top_z = 0.5 * length
    bottom_z = -top_z
    tolerance = 1e-14 * max(1.0, min(top_z, radius))
    cylinder_class = CylinderClass.Medium
    if top_z - radius > tolerance:
        cylinder_class = CylinderClass.Long
    elif radius - top_z > tolerance:
        cylinder_class = CylinderClass.Short

    n_vertices_per_circle = max(3, math.ceil(2.0 * np.pi * radius / resolution_hint))

    mesh_vertices = []

    bottom_center = len(mesh_vertices)
    mesh_vertices.append(np.array([0.0, 0.0, bottom_z]))
    top_center = len(mesh_vertices)
    mesh_vertices.append(np.array([0.0, 0.0, top_z]))

    bottom = []
    top = []
    angle_step = 2.0 * np.pi / n_vertices_per_circle
    for i in range(n_vertices_per_circle):
        x = radius * np.cos(angle_step * i)
        y = radius * np.sin(angle_step * i)
        bottom.append(len(mesh_vertices))
        mesh_vertices.append(np.array([x, y, bottom_z]))
        top.append(len(mesh_vertices))
        mesh_vertices.append(np.array([x, y, top_z]))

    n_outer_vertices = len(mesh_vertices)

    potentials = [0.0] * n_outer_vertices

    if cylinder_class == CylinderClass.Long:
        mesh_elements = _calc_long_cylinder_volume_mesh_with_ma(
            radius, length, n_vertices_per_circle, bottom_center, bottom,
            top_center, top, mesh_vertices, potentials)
    elif cylinder_class == CylinderClass.Medium:
        mesh_elements = _calc_medium_cylinder_volume_mesh_with_ma(
            radius, n_vertices_per_circle, bottom_center, bottom, top_center,
            top, mesh_vertices, potentials)
    else:
        assert cylinder_class == CylinderClass.Short
        mesh_elements = _calc_short_cylinder_volume_mesh_with_ma(
            radius, length, n_vertices_per_circle, bottom_center, bottom,
            top_center, top, mesh_vertices, potentials)

    return (np.array(mesh_vertices), np.array(mesh_elements, dtype=int),
            np.array(potentials))


class CylinderClass(enum.Enum):
    Long = 0
    Medium = 1
    Short = 2


def _calc_long_cylinder_volume_mesh_with_ma(
        radius, length, n_vertices_per_circle, bottom_center, bottom,
        top_center, top, mesh_vertices, potentials):
    medial = []
    offset_distance = radius
    top_z = 0.5 * length
    offset_top_z = top_z - offset_distance
    offset_bottom_z = -offset_top_z
    medial.append(len(mesh_vertices))
    mesh_vertices.append(np.array([0.0, 0.0, offset_bottom_z]))
    potentials.append(radius)
    medial.append(len(mesh_vertices))
    mesh_vertices.append(np.array([0.0, 0.0, offset_top_z]))
    potentials.append(radius)

    mesh_elements = []
    i = n_vertices_per_circle - 1
    for j in range(n_vertices_per_circle):
        mesh_elements.append([bottom_center, bottom[i], bottom[j], medial[0]])
        mesh_elements.append([top_center, top[j], top[i], medial[1]])
        mesh_elements.extend(_split_triangular_prism_to_tetrahedra(
            medial[0], bottom[i], bottom[j], medial[1], top[i], top[j]))
        i = j

    return mesh_elements


def _calc_medium_cylinder_volume_mesh_with_ma(
        radius, n_vertices_per_circle, bottom_center, bottom, top_center, top,
        mesh_vertices, potentials):
    medial = len(mesh_vertices)
    mesh_vertices.append(np.array([0.0, 0.0, 0.0]))
    potentials.append(radius)

    mesh_elements = []
    i = n_vertices_per_circle - 1
    for j in range(n_vertices_per_circle):
        mesh_elements.append([bottom_center, bottom[i], bottom[j], medial])
        mesh_elements.append([top_center, top[j], top[i], medial])
        mesh_elements.extend(_split_pyramid_to_tetrahedra(
            top[i], top[j], bottom[j], bottom[i], medial))
        i = j

    return mesh_elements


def _calc_short_cylinder_volume_mesh_with_ma(
        radius, length, n_vertices_per_circle, bottom_center, bottom,
        top_center, top, mesh_vertices, potentials):
    center = len(mesh_vertices)
    mesh_vertices.append(np.array([0.0, 0.0, 0.0]))
    half_length = 0.5 * length
    potentials.append(half_length)

    medial = []
    medial_radius = radius - half_length
    scale_cylinder_radius_to_medial_circle = medial_radius / radius
    for i in range(n_vertices_per_circle):
        x = mesh_vertices[bottom[i]][0] * scale_cylinder_radius_to_medial_circle
        y = mesh_vertices[bottom[i]][1] * scale_cylinder_radius_to_medial_circle
        medial.append(len(mesh_vertices))
        mesh_vertices.append(np.array([x, y, 0.0]))
        potentials.append(half_length)

    mesh_elements = []
    i = n_vertices_per_circle - 1
    for j in range(n_vertices_per_circle):
        mesh_elements.extend(_split_triangular_prism_to_tetrahedra(
            bottom_center, bottom[i], bottom[j], center, medial[i], medial[j]))
        mesh_elements.extend(_split_triangular_prism_to_tetrahedra(
            center, medial[i], medial[j], top_center, top[i], top[j]))
        mesh_elements.extend(_split_triangular_prism_to_tetrahedra(
            bottom[i], medial[i], top[i], bottom[j], medial[j], top[j]))
        i = j

    return mesh_elements


def _split_triangular_prism_to_tetrahedra(v0, v1, v2, v3, v4, v5):
    elements = []
    previous = v3
    for next in [v4, v1, v2]:
        elements.append([previous, next, v0, v5])
        previous = next
    return elements


def _split_pyramid_to_tetrahedra(v0, v1, v2, v3, v4):
    elements = []
    previous = v3
    for next in [v4, v1]:
        elements.append([previous, next, v0, v2])
        previous = next
    return elements


def make_tetrahedral_capsule(radius, height, resolution_hint):
    """Create a tetrahedral capsule mesh.

    Source: Drake (https://github.com/RobotLocomotion/drake/blob/903019faf53b771e8e2fe81222ffd74eae2dc85c/geometry/proximity/make_capsule_mesh.cc#L14),
    BSD 3-clause

    Parameters
    ----------
    radius : float
        Radius of the capsule.

    height : float
        Height of the capsule.

    resolution_hint : float
        Controls the fineness of the tetrahedral mesh. The coarsest mesh
        that produces desirable results will allow simulation to run as
        efficiently as possible. The circles of the cylinder and great circles
        of each hemisphere will have 2 * pi * radius / resolution_hint edges.

    Returns
    -------
    vertices : array, shape (n_vertices, 3)
        Vertices of the mesh.

    tetrahedra : array, shape (n_tetrahedra, 4)
        Indices of vertices that form tetrahedra of the mesh.

    potentials : array, shape (n_vertices, 3)
        Potential of each vertex.
    """
    medial_top_z = 0.5 * height
    medial_bottom_z = -medial_top_z
    top_z = medial_top_z + radius
    bottom_z = -top_z

    n_vertices_per_circle = int(np.clip(2.0 * np.pi * radius / resolution_hint,
                                        3.0, 706.0))
    n_circles_per_cap = n_vertices_per_circle // 2

    mesh_vertices = []

    medial_top = len(mesh_vertices)
    mesh_vertices.append(np.array([0.0, 0.0, medial_top_z]))
    medial_bottom = len(mesh_vertices)
    mesh_vertices.append(np.array([0.0, 0.0, medial_bottom_z]))
    top = len(mesh_vertices)
    mesh_vertices.append(np.array([0.0, 0.0, top_z]))
    bottom = len(mesh_vertices)
    mesh_vertices.append(np.array([0.0, 0.0, bottom_z]))

    top_cap = []
    bottom_cap = []

    theta_step = 0.5 * np.pi / n_circles_per_cap
    phi_step = 2.0 * np.pi / n_vertices_per_circle

    for i in range(n_circles_per_cap):
        theta = 0.5 * np.pi - i * theta_step
        s = np.sin(theta)
        top_circle_z = radius * np.cos(theta) + medial_top_z
        bottom_circle_z = -top_circle_z
        for j in range(n_vertices_per_circle):
            phi = j * phi_step
            x = radius * s * np.cos(phi)
            y = radius * s * np.sin(phi)

            top_cap.append(len(mesh_vertices))
            mesh_vertices.append(np.array([x, y, top_circle_z]))
            bottom_cap.append(len(mesh_vertices))
            mesh_vertices.append(np.array([x, y, bottom_circle_z]))

    mesh_elements = []
    for i in range(n_circles_per_cap - 1):
        for j in range(n_vertices_per_circle):
            j1 = (j + 1) % n_vertices_per_circle
            mesh_elements.extend(_split_pyramid_to_tetrahedra(
                top_cap[(i + 1) * n_vertices_per_circle + j],
                top_cap[(i + 1) * n_vertices_per_circle + j1],
                top_cap[i * n_vertices_per_circle + j1],
                top_cap[i * n_vertices_per_circle + j],
                medial_top
            ))
            mesh_elements.extend(_split_pyramid_to_tetrahedra(
                bottom_cap[i * n_vertices_per_circle + j],
                bottom_cap[i * n_vertices_per_circle + j1],
                bottom_cap[(i + 1) * n_vertices_per_circle + j1],
                bottom_cap[(i + 1) * n_vertices_per_circle + j],
                medial_bottom
            ))

    last_circle_offset = (n_circles_per_cap - 1) * n_vertices_per_circle
    for j in range(n_vertices_per_circle):
        j1 = (j + 1) % n_vertices_per_circle
        mesh_elements.append([top, top_cap[last_circle_offset + j1],
                              top_cap[last_circle_offset + j], medial_top])
        mesh_elements.append([bottom, bottom_cap[last_circle_offset + j],
                              bottom_cap[last_circle_offset + j1],
                              medial_bottom])
        mesh_elements.extend(_split_triangular_prism_to_tetrahedra(
            medial_bottom, bottom_cap[j], bottom_cap[j1], medial_top,
            top_cap[j], top_cap[j1]))

    potentials = np.zeros(len(mesh_vertices))
    potentials[:2] = radius

    return (np.array(mesh_vertices), np.array(mesh_elements, dtype=int),
            potentials)



"""Hydroelastic contact model for contact wrenches of rigid bodies.

The original name for this model is pressure field model. It has been
presented in:

R. Elandt, E. Drumwright, M. Sherman and A. Ruina: A pressure field model for
fast, robust approximation of net contact force and moment between nominally
rigid objects, IEEE/RSJ International Conference on Intelligent Robots and
Systems (IROS), 2019, pp. 8238-8245, doi: 10.1109/IROS40897.2019.8968548.

The implementation is different, but is inspired by:

V. Huang, F. Wang, E. Zhang: An Efficient Implementation of Pressure Field
Models, https://www.ekzhang.com/assets/pdf/Hydroelastics.pdf.
"""
from ._interface import contact_forces, find_contact_surface
from ._rigid_body import RigidBody
from ._contact_surface import ContactSurface

from ._halfplanes import (
    intersect_halfplanes, plot_halfplanes_and_intersections)
from ._mesh_processing import (
    center_of_mass_tetrahedral_mesh, tetrahedral_mesh_aabbs,
    tetrahedral_mesh_volumes)
from ._barycentric_transform import barycentric_transforms
from ._tetrahedron_intersection import intersect_tetrahedron_pair
from ._forces import compute_contact_force
from ._broad_phase import HydroelasticBoundingVolumeHierarchy


__all__ = [
    "contact_forces", "find_contact_surface", "RigidBody", "ContactSurface",
    "HydroelasticBoundingVolumeHierarchy",
    # exported for unit tests and specific examples:
    "intersect_halfplanes", "plot_halfplanes_and_intersections",
    "center_of_mass_tetrahedral_mesh", "tetrahedral_mesh_aabbs",
    "tetrahedral_mesh_volumes", "barycentric_transforms",
    "intersect_tetrahedron_pair", "compute_contact_force",
]



"""Minkowski portal refinement (MPR) for collision detection.

Another name for the algorithm is XenoCollide (for details, see
http://xenocollide.snethen.com/). It has been presented in "Game Programming
Gems 7" (Gem 2.5: XenoCollide: Complex Collision Made Simple).

This implementation of MPR is based on libccd (for details, see
https://github.com/danfis/libccd). For the original code the copyright is of
Daniel Fiser <danfis@danfis.cz>. It has been released under 3-clause BSD
license.
"""
from enum import Enum
import numpy as np
import numba

from .minkowski import Simplex, support_function, make_support_point
from .utils import norm_vector, EPSILON
from .distance import point_to_triangle


def mpr_intersection(collider1, collider2, mpr_tolerance=0.0001):
    """Intersection test with Minkowski Portal Refinement (MPR).

    Parameters
    ----------
    collider1 : ConvexCollider
        Convex collider 1.

    collider2 : ConvexCollider
        Convex collider 2.

    mpr_tolerance : float
        Boundary tolerance for MPR algorithm

    Returns
    -------
    intersection : bool
        Do the two colliders intersect?
    """
    res, portal = _discover_portal(collider1, collider2)
    if res == PortalState.ORIGIN_OUTSIDE_PORTAL:
        return False
    elif res == PortalState.PORTAL_WAS_BUILT:
        return _refine_portal(collider1, collider2, portal, mpr_tolerance)
    else:
        return True


def mpr_penetration(collider1, collider2, mpr_tolerance=0.0001, max_iterations=100):
    """Minkowski Portal Refinement (MPR) with penetration info.

    The returned penetration direction and contact position form a line along
    which the penetration happens. The penetration direction is computed such
    that adding depth + penetration_direction to each vertex of the second
    collider separates the two colliders.

    Parameters
    ----------
    collider1 : ConvexCollider
        Convex collider 1.

    collider2 : ConvexCollider
        Convex collider 2.

    mpr_tolerance : float, optional (default: 0.0001)
        Boundary tolerance for MPR algorithm

    max_iterations : int, optional (default: 100)
        Maximum number of iterations to compute penetration info.

    Returns
    -------
    intersection : bool
        Do the two colliders intersect?

    depth : float or None
        Penetration depth.

    penetration_direction : array, shape (3,) or None
        Penetration direction. Vector will be either a unit vector or zero
        vector in case of touching colliders.

    contact_position : array, shape (3,) or None
        Contact position.
    """
    res, portal = _discover_portal(collider1, collider2)
    depth, penetration_direction, contact_position = None, None, None
    if res == PortalState.ORIGIN_OUTSIDE_PORTAL:
        intersection = False
    elif res == PortalState.ORIGIN_ON_V1:
        intersection = True
        depth, penetration_direction, contact_position = _find_penetration_touch(
            portal.v1, portal.v2)
    elif res == PortalState.ORIGIN_ON_V0V1_SEGMENT:
        intersection = True
        depth, penetration_direction, contact_position = _find_penetration_segment(
            portal.v, portal.v1, portal.v2)
    else:
        assert res == PortalState.PORTAL_WAS_BUILT
        intersection = _refine_portal(collider1, collider2, portal, mpr_tolerance)
        if intersection:
            depth, penetration_direction, contact_position = _find_penetration_info(
                collider1, collider2, portal, mpr_tolerance, max_iterations)
    return intersection, depth, penetration_direction, contact_position


class PortalState(Enum):
    CONTINUE_BUILDING_PORTAL = -2
    ORIGIN_OUTSIDE_PORTAL = -1
    PORTAL_WAS_BUILT = 0
    ORIGIN_ON_V1 = 1
    ORIGIN_ON_V0V1_SEGMENT = 2


def _discover_portal(collider1, collider2):
    """Discover initial portal.

    The initial portal is a tetrahedron that intersects with the ray from
    the center of the Minkowski difference to the origin.
    """
    portal = Simplex()
    _find_origin_ray(portal, collider1, collider2)

    origin_outside_v1 = _find_support_in_direction_of_origin_ray(
        portal, collider1, collider2)
    if origin_outside_v1:
        return PortalState.ORIGIN_OUTSIDE_PORTAL, portal

    res = _find_support_perpendicular_to_plane_containing_origin_v01(
        portal, collider1, collider2)
    if res != PortalState.CONTINUE_BUILDING_PORTAL:
        return res, portal
    portal.n_points = 3

    search_direction = _search_direction_perpendicular_to_plane_containing_v012(
        portal.v, portal.v1, portal.v2)
    while portal.n_points < 4:
        portal.v[3], portal.v1[3], portal.v2[3] = support_function(
            collider1, collider2, search_direction)
        if portal.v[3].dot(search_direction) < EPSILON:
            return PortalState.ORIGIN_OUTSIDE_PORTAL, portal
        search_direction, portal.n_points = _iterate_discover_portal(
            portal.v, portal.v1, portal.v2, search_direction, portal.n_points)

    return PortalState.PORTAL_WAS_BUILT, portal


def _find_origin_ray(portal, collider1, collider2):
    """Find origin ray.

    The origin ray starts at v0 and passes through the origin. v0 must be a
    point that is known to be in the interior of the Minkowski difference.
    Such a point can be obtained from points inside the colliders. Center
    points are good candidates.
    """
    portal.v[0], portal.v1[0], portal.v2[0] = make_support_point(
        collider1.center(), collider2.center())
    portal.n_points = 1
    portals_center_is_origin = all(portal.v[0] == 0.0)
    if portals_center_is_origin:
        # Colliders intersect, but we need penetration info.
        portal.v[0, 0] += EPSILON * 10.0


def _find_support_in_direction_of_origin_ray(portal, collider1, collider2):
    search_direction = norm_vector(-portal.v[0])
    portal.v[1], portal.v1[1], portal.v2[1] = support_function(
        collider1, collider2, search_direction)
    portal.n_points = 2
    origin_outside_v1 = (
        any(portal.v[1] != 0.0) and
        portal.v[1].dot(search_direction) < EPSILON)
    return origin_outside_v1


def _find_support_perpendicular_to_plane_containing_origin_v01(
        portal, collider1, collider2):
    search_direction = np.cross(portal.v[0], portal.v[1])
    if search_direction.dot(search_direction) < EPSILON:
        if all(portal.v[1] == 0.0):
            return PortalState.ORIGIN_ON_V1
        else:
            return PortalState.ORIGIN_ON_V0V1_SEGMENT
    search_direction = norm_vector(search_direction)
    portal.v[2], portal.v1[2], portal.v2[2] = support_function(
        collider1, collider2, search_direction)
    if portal.v[2].dot(search_direction) < EPSILON:
        return PortalState.ORIGIN_OUTSIDE_PORTAL
    return PortalState.CONTINUE_BUILDING_PORTAL


@numba.njit(cache=True)
def _search_direction_perpendicular_to_plane_containing_v012(v, v1, v2):
    search_direction = norm_vector(np.cross(v[1] - v[0], v[2] - v[0]))
    if search_direction.dot(v[0]) > 0.0:
        _swap_vertices(v, v1, v2, 1, 2)
        search_direction *= -1.0
    return search_direction


@numba.njit(cache=True)
def _iterate_discover_portal(v, v1, v2, search_direction, portal_size):
    cont = False
    origin_outside_v103 = np.cross(v[1], v[3]).dot(v[0]) < EPSILON
    if origin_outside_v103:
        v[2], v1[2], v2[2] = v[3], v1[3], v2[3]
        cont = True
    if not cont:
        origin_outside_v302 = np.cross(v[3], v[2]).dot(v[0]) < EPSILON
        if origin_outside_v302:
            v[1], v1[1], v2[1] = v[3], v1[3], v2[3]
            cont = True
    if cont:
        search_direction = norm_vector(np.cross(v[1] - v[0], v[2] - v[0]))
    else:
        portal_size = 4
    return search_direction, portal_size


@numba.njit(cache=True)
def _swap_vertices(v, v1, v2, idx1, idx2):
    tmp = v[idx1]
    tmp1 = v1[idx1]
    tmp2 = v2[idx1]
    v[idx1] = v[idx2]
    v1[idx1] = v1[idx2]
    v2[idx1] = v2[idx2]
    v[idx2] = tmp
    v1[idx2] = tmp1
    v2[idx2] = tmp2


def _refine_portal(collider1, collider2, portal, mpr_tolerance):
    """Expands portal towards origin and determine if objects intersect.

    Parameters
    ----------
    portal : Simplex
        Already established portal.

    Returns
    -------
    intersection_found : bool
        Was an intersection found?
    """
    while True:
        search_direction = _portal_direction(portal.v)
        if _encapsulates_origin(portal.v[1], search_direction):
            return True
        next_support_point, next_support_point1, next_support_point2 = support_function(
            collider1, collider2, search_direction)
        if (not _encapsulates_origin(next_support_point, search_direction)
                or _portal_reach_tolerance(portal.v, next_support_point,
                                           search_direction, mpr_tolerance)):
            return False
        _expand_portal(portal.v, portal.v1, portal.v2, next_support_point,
                       next_support_point1, next_support_point2)


@numba.njit(cache=True)
def _portal_direction(v):
    """Compute direction outside portal (from v0 through v1-v2-v3 face).

    Portal's v1-v2-v3 face must be arranged in correct order!
    """
    return norm_vector(np.cross(v[2] - v[1], v[3] - v[1]))


@numba.njit(cache=True)
def _encapsulates_origin(v, search_direction):
    """Does the portal encapsulate the origin?"""
    return v.dot(search_direction) > -10.0 * EPSILON


@numba.njit(cache=True)
def _portal_reach_tolerance(v, v4, search_direction, mpr_tolerance):
    """Returns if portal with new point v4 would reach a specified tolerance.

    Returns true if portal can't significantly expand within Minkowski
    difference. v4 is candidate for new point in portal, search_direction is
    direction in which v4 was obtained.
    """
    return min(v4.dot(search_direction) - v[1:].dot(search_direction)) < mpr_tolerance + EPSILON


@numba.njit(cache=True)
def _expand_portal(v, v1, v2, v4, v14, v24):
    """Extends portal with new support point.

    Portal must have face v1-v2-v3 arranged to face outside portal.
    """
    v4v0 = np.cross(v4, v[0])
    if v[1].dot(v4v0) > 0.0:
        if v[2].dot(v4v0) > 0.0:
            v[1], v1[1], v2[1] = v4, v14, v24
        else:
            v[3], v1[3], v2[3] = v4, v14, v24
    else:
        if v[3].dot(v4v0) > 0.0:
            v[2], v1[2], v2[2] = v4, v14, v24
        else:
            v[1], v1[1], v2[1] = v4, v14, v24


def _find_penetration_info(collider1, collider2, portal, mpr_tolerance, max_iterations):
    """Finds penetration info by expanding the portal."""
    iterations = 0
    while True:
        search_direction = _portal_direction(portal.v)
        next_support_point, next_support_point1, next_support_point2 = support_function(
            collider1, collider2, search_direction)
        tolerance_reached = (
            _portal_reach_tolerance(
                portal.v, next_support_point, search_direction, mpr_tolerance)
            or iterations > max_iterations)
        if tolerance_reached:
            depth, pdir, pos = _penetration_info(portal.v, portal.v1, portal.v2)
            return depth, norm_vector(pdir), pos
        _expand_portal(portal.v, portal.v1, portal.v2, next_support_point,
                       next_support_point1, next_support_point2)
        iterations += 1


@numba.njit(cache=True)
def _penetration_info(v, v1, v2):
    depth, penetration_direction = point_to_triangle(np.zeros(3), v[1:])
    if abs(depth) < EPSILON:
        # Touching contact. Direction does not matter.
        penetration_direction = np.zeros(3)
    return depth, penetration_direction, _contact_position(
        v, v1, v2, _portal_direction(v))


@numba.njit(cache=True)
def _find_penetration_touch(v1, v2):
    """Finds penetration info if origin lies on v1 (touching contact)."""
    depth = 0.0
    penetration_direction = np.zeros(3)
    contact_position = 0.5 * (v1[1] + v2[1])
    return depth, penetration_direction, contact_position


@numba.njit(cache=True)
def _find_penetration_segment(v, v1, v2):
    """Find penetration info if origin lies on segment v0-v1.

    Depth is distance to v1. Direction also and position must be computed.
    """
    contact_position = 0.5 * (v1[1] + v2[1])
    penetration_direction = v[1]
    depth = np.linalg.norm(penetration_direction)
    return depth, norm_vector(penetration_direction), contact_position


@numba.njit(cache=True)
def _contact_position(v, v1, v2, search_direction):
    """Compute contact position from fully established portal."""
    barycentric_coordinates = np.empty(4)
    barycentric_coordinates[0] = np.cross(v[1], v[2]).dot(v[3])
    barycentric_coordinates[1] = np.cross(v[3], v[2]).dot(v[0])
    barycentric_coordinates[2] = np.cross(v[0], v[1]).dot(v[3])
    barycentric_coordinates[3] = np.cross(v[2], v[1]).dot(v[0])

    coords_sum = np.sum(barycentric_coordinates)

    if coords_sum < EPSILON:
        barycentric_coordinates = np.array([
            0.0,
            np.cross(v[2], v[3]).dot(search_direction),
            np.cross(v[3], v[1]).dot(search_direction),
            np.cross(v[1], v[2]).dot(search_direction)
        ])
        coords_sum = np.sum(barycentric_coordinates)

    barycentric_coordinates /= coords_sum

    v1 = barycentric_coordinates.dot(v1)
    v2 = barycentric_coordinates.dot(v2)

    return 0.5 * (v1 + v2)



from ._forces import contact_surface_forces, accumulate_wrenches
from ._barycentric_transform import barycentric_transforms
from ._tetrahedron_intersection import intersect_tetrahedron_pairs
from ._contact_surface import ContactSurface
from ..aabb_tree import all_aabbs_overlap


def contact_forces(rigid_body1, rigid_body2, return_details=False):
    """Contact forces between two objects.

    Parameters
    ----------
    rigid_body1 : RigidBody
        First rigid body.

    rigid_body2 : RigidBody
        Second rigid body.

    return_details : bool
        Return additional contact details.

    Returns
    -------
    intersection : bool
        Do both bodies intersect?

    wrench12_in_world : array, shape (6,)
        Forces and torques caused by body 1 acting on body 2 in world frame.

    wrench21_in_world : array, shape (6,)
        Forces and torques caused by body 2 acting on body 1 in world frame.

    details : dict, optional
        Additional contact details.
    """
    contact_surface = find_contact_surface(rigid_body1, rigid_body2)

    wrench12_in_world, wrench21_in_world = accumulate_wrenches(
        contact_surface, rigid_body1, rigid_body2)

    if return_details:
        details = contact_surface.make_details(
            rigid_body1.tetrahedra_points, rigid_body2.tetrahedra_points)
        return (
            contact_surface.intersection, wrench12_in_world, wrench21_in_world,
            details)
    else:
        return (
            contact_surface.intersection, wrench12_in_world, wrench21_in_world)


def find_contact_surface(rigid_body1, rigid_body2, use_aabb_trees=False):
    """Find contact plane of two rigid bodies.

    Note that this function will transform rigid_body1 into the frame of
    rigid_body2.

    Parameters
    ----------
    rigid_body1 : RigidBody
        First rigid body.

    rigid_body2 : RigidBody
        Second rigid body.

    use_aabb_trees : bool, optional (default: False)
        Option to specify the usage of aabb_trees.

    Returns
    -------
    contact_surface : ContactSurface
        Contact information.
    """
    # We transform vertices of rigid_body1 to rigid_body2 frame to be able to
    # reuse the AABB tree of rigid_body2.
    rigid_body1.express_in(rigid_body2.body2origin_)

    if use_aabb_trees:
        _, broad_tetrahedra1, broad_tetrahedra2, broad_pairs \
            = rigid_body1.aabbtree_.overlaps_aabb_tree(rigid_body2.aabbtree_)
    else:
        broad_tetrahedra1, broad_tetrahedra2, broad_pairs = all_aabbs_overlap(rigid_body1.aabbs, rigid_body2.aabbs)

    X1 = barycentric_transforms(rigid_body1.tetrahedra_points[broad_tetrahedra1])
    X2 = barycentric_transforms(rigid_body2.tetrahedra_points[broad_tetrahedra2])
    X1 = {j: X1[i] for i, j in enumerate(broad_tetrahedra1)}
    X2 = {j: X2[i] for i, j in enumerate(broad_tetrahedra2)}

    intersection_result = intersect_tetrahedron_pairs(
        broad_pairs,
        rigid_body1.tetrahedra_points, rigid_body2.tetrahedra_points,
        rigid_body1.tetrahedra_potentials, rigid_body2.tetrahedra_potentials,
        X1, X2, rigid_body1.youngs_modulus, rigid_body2.youngs_modulus)
    contact_surface = ContactSurface(
        rigid_body2.body2origin_, *intersection_result)

    areas, coms, forces, triangles = contact_surface_forces(
        contact_surface, rigid_body1)
    contact_surface.add_polygon_info(areas, coms, forces, triangles)

    return contact_surface



"""Benchmark utilities."""
import time


class Timer:
    """Collect timing information from benchmarks."""
    def __init__(self):
        self._start_time = {}
        self.total_time_ = {}

    def start(self, name):
        self._start_time[name] = time.time()

    def stop(self, name):
        stop_time = time.time()
        assert name in self._start_time
        return stop_time - self._start_time[name]

    def stop_and_add_to_total(self, name):
        duration = self.stop(name)
        current_total = self.total_time_.get(name, 0.0)
        self.total_time_[name] = current_total + duration
